from abc import abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Generic, Literal, TypeAlias, Union

from pydantic import BaseModel
from typing_extensions import TypeVar

from dagster import _check as check
from dagster._annotations import public

# Constant for object attribute name
SCAFFOLDER_CLS_ATTRIBUTE: str = "__scaffolder_cls__"


class NoParams(BaseModel): ...


# Type variable for generic class handling
T = TypeVar("T")
TModel = TypeVar("TModel", bound=BaseModel, default=NoParams)


@public
def scaffold_with(
    scaffolder_cls: Union[type["Scaffolder[Any]"], "ScaffolderUnavailableReason"],
) -> Callable[[T], T]:
    """A decorator that declares what :py:class:`Scaffolder` is used to scaffold the artifact.

    Args:
        scaffolder_cls: A class that inherits from Scaffolder

    Returns:
        Decorator function that enhances the target class

    Example:
        .. code-block:: python
            import dagster as dg

            class MyArtifactScaffolder(dg.Scaffolder):
                ...

            @dg.scaffold_with(MyArtifactScaffolder)
            class MyArtifact:
                ...

    """
    from dagster.components.core.package_entry import PACKAGE_ENTRY_ATTR

    def decorator(obj: T) -> T:
        # Store the scaffolder class as an attribute using the constant
        setattr(obj, SCAFFOLDER_CLS_ATTRIBUTE, scaffolder_cls)
        # All scaffoldable objects are registry objects
        setattr(obj, PACKAGE_ENTRY_ATTR, True)
        return obj

    return decorator


def has_scaffolder(obj: object) -> bool:
    """Determines if an object has been decorated with `@scaffold_with`.

    Args:
        obj: The object to check

    Returns:
        True if the object has a Scaffolder attached, False otherwise
    """
    return hasattr(obj, SCAFFOLDER_CLS_ATTRIBUTE)


def get_scaffolder(obj: object) -> Union["Scaffolder[Any]", "ScaffolderUnavailableReason"]:
    """Retrieves the scaffolder class attached to the decorated object.

    Args:
        obj: The object to inspect

    Returns:
        The scaffolder class attached to the decorated object. Raises CheckError if the object is not decorated with @scaffold_with.
    """
    check.param_invariant(
        has_scaffolder(obj), "obj", "Object must be decorated with @scaffold_with"
    )
    attr = getattr(obj, SCAFFOLDER_CLS_ATTRIBUTE)
    return attr if isinstance(attr, ScaffolderUnavailableReason) else attr()


@dataclass
class ScaffolderUnavailableReason:
    message: str


ScaffoldFormatOptions: TypeAlias = Literal["yaml", "python"]


@public
@dataclass
class ScaffoldRequest(Generic[TModel]):
    """Details about the current scaffolding operation.

    This is passed to the :py:class:`Scaffolder` class when scaffolding a target.
    """

    # fully qualified class name of the decorated object
    type_name: str
    # target path for the scaffold request. Typically used to construct absolute paths
    target_path: Path
    # yaml or python
    scaffold_format: ScaffoldFormatOptions
    # the root of the dg project
    project_root: Path | None
    # optional params for scaffolding
    params: TModel
    # whether to append to an existing file
    append: bool = False


@public
class Scaffolder(Generic[TModel]):
    """Handles scaffolding its associated scaffold target based on user-supplied parameters.
    Invoked by a user using the `dg scaffold` CLI command.

    To associate a scaffolder with its target class, use the :py:func:`scaffold_with` decorator.
    """

    @classmethod
    def get_scaffold_params(cls) -> type[TModel]:
        """Returns the model class that contains the parameters for scaffolding. By default,
        this is :py:class:`NoParams`, indicating that no additional parameters can be supplied
        to the scaffolder.
        """
        return NoParams  # type: ignore

    @abstractmethod
    def scaffold(self, request: ScaffoldRequest[TModel]) -> None:
        """Scaffold the target with the given request.

        Args:
            request: The scaffold request containing type name, target path, format, project root and params.
                The params are validated against the model returned by :py:meth:`get_scaffold_params`.
        """
        ...



import copy
from collections.abc import Callable, Mapping, Sequence
from datetime import datetime
from typing import TYPE_CHECKING, Any, Generic, TypeAlias, Union, cast

from typing_extensions import TypeVar

import dagster._check as check
from dagster._annotations import deprecated, deprecated_param, public
from dagster._core.definitions.config import ConfigMapping
from dagster._core.definitions.partitions.definition.partitions_definition import (
    PartitionsDefinition,
)
from dagster._core.definitions.partitions.partition import Partition
from dagster._core.errors import DagsterInvalidInvocationError
from dagster._core.storage.tags import PARTITION_SET_TAG
from dagster._utils import xor
from dagster._utils.tags import normalize_tags

if TYPE_CHECKING:
    from dagster._core.definitions.run_config import RunConfig

PartitionConfigFn: TypeAlias = "Callable[[str], RunConfig | Mapping[str, Any]]"
T_PartitionsDefinition = TypeVar(
    "T_PartitionsDefinition",
    bound="PartitionsDefinition",
    default="PartitionsDefinition",
    covariant=True,
)


@public
@deprecated_param(
    param="run_config_for_partition_fn",
    breaking_version="2.0",
    additional_warn_text="Use `run_config_for_partition_key_fn` instead.",
)
@deprecated_param(
    param="tags_for_partition_fn",
    breaking_version="2.0",
    additional_warn_text="Use `tags_for_partition_key_fn` instead.",
)
class PartitionedConfig(Generic[T_PartitionsDefinition]):
    """Defines a way of configuring a job where the job can be run on one of a discrete set of
    partitions, and each partition corresponds to run configuration for the job.

    Setting PartitionedConfig as the config for a job allows you to launch backfills for that job
    and view the run history across partitions.
    """

    def __init__(
        self,
        partitions_def: T_PartitionsDefinition,
        run_config_for_partition_fn: Callable[[Partition], Mapping[str, Any]] | None = None,
        decorated_fn: Callable[..., Union["RunConfig", Mapping[str, Any]]] | None = None,
        tags_for_partition_fn: Callable[[Partition[Any]], Mapping[str, str]] | None = None,
        run_config_for_partition_key_fn: "PartitionConfigFn | None" = None,
        tags_for_partition_key_fn: Callable[[str], Mapping[str, str]] | None = None,
    ):
        self._partitions = check.inst_param(partitions_def, "partitions_def", PartitionsDefinition)
        self._decorated_fn = decorated_fn

        check.invariant(
            xor(run_config_for_partition_fn, run_config_for_partition_key_fn),
            "Must provide exactly one of run_config_for_partition_fn or"
            " run_config_for_partition_key_fn",
        )
        check.invariant(
            not (tags_for_partition_fn and tags_for_partition_key_fn),
            "Cannot provide both of tags_for_partition_fn or tags_for_partition_key_fn",
        )

        self._run_config_for_partition_fn = check.opt_callable_param(
            run_config_for_partition_fn, "run_config_for_partition_fn"
        )
        self._run_config_for_partition_key_fn = check.opt_callable_param(
            run_config_for_partition_key_fn, "run_config_for_partition_key_fn"
        )
        self._tags_for_partition_fn = check.opt_callable_param(
            tags_for_partition_fn, "tags_for_partition_fn"
        )
        self._tags_for_partition_key_fn = check.opt_callable_param(
            tags_for_partition_key_fn, "tags_for_partition_key_fn"
        )

    @public
    @property
    def partitions_def(
        self,
    ) -> T_PartitionsDefinition:
        """T_PartitionsDefinition: The partitions definition associated with this PartitionedConfig."""
        return self._partitions

    @deprecated(
        breaking_version="2.0",
        additional_warn_text="Use `run_config_for_partition_key_fn` instead.",
    )
    @public
    @property
    def run_config_for_partition_fn(
        self,
    ) -> Callable[[Partition], Mapping[str, Any]] | None:
        """Optional[Callable[[Partition], Mapping[str, Any]]]: A function that accepts a partition
        and returns a dictionary representing the config to attach to runs for that partition.
        Deprecated as of 1.3.3.
        """
        return self._run_config_for_partition_fn

    @public
    @property
    def run_config_for_partition_key_fn(
        self,
    ) -> "PartitionConfigFn | None":
        """Optional[Callable[[str], Union[RunConfig, Mapping[str, Any]]]]: A function that accepts a partition key
        and returns a dictionary representing the config to attach to runs for that partition.
        """
        return self._run_config_for_partition_key_fn

    @deprecated(
        breaking_version="2.0", additional_warn_text="Use `tags_for_partition_key_fn` instead."
    )
    @public
    @property
    def tags_for_partition_fn(self) -> Callable[[Partition], Mapping[str, str]] | None:
        """Optional[Callable[[Partition], Mapping[str, str]]]: A function that
        accepts a partition and returns a dictionary of tags to attach to runs for
        that partition. Deprecated as of 1.3.3.
        """
        return self._tags_for_partition_fn

    @public
    @property
    def tags_for_partition_key_fn(
        self,
    ) -> Callable[[str], Mapping[str, str]] | None:
        """Optional[Callable[[str], Mapping[str, str]]]: A function that
        accepts a partition key and returns a dictionary of tags to attach to runs for
        that partition.
        """
        return self._tags_for_partition_key_fn

    @public
    def get_partition_keys(self, current_time: datetime | None = None) -> Sequence[str]:
        """Returns a list of partition keys, representing the full set of partitions that
        config can be applied to.

        Args:
            current_time (Optional[datetime]): A datetime object representing the current time. Only
                applicable to time-based partitions definitions.

        Returns:
            Sequence[str]
        """
        return self.partitions_def.get_partition_keys(current_time)

    # Assumes partition key already validated
    def get_run_config_for_partition_key(
        self,
        partition_key: str,
    ) -> Mapping[str, Any]:
        """Generates the run config corresponding to a partition key.

        Args:
            partition_key (str): the key for a partition that should be used to generate a run config.
        """
        from dagster._core.definitions.run_config import convert_config_input

        # _run_config_for_partition_fn is deprecated, we can remove this branching logic in 2.0
        if self._run_config_for_partition_fn:
            run_config = self._run_config_for_partition_fn(Partition(partition_key))
        elif self._run_config_for_partition_key_fn:
            run_config = self._run_config_for_partition_key_fn(partition_key)
        else:
            check.failed("Unreachable.")  # one of the above funcs always defined
        normalized_run_config = convert_config_input(
            run_config
        )  # convert any RunConfig to plain dict
        return copy.deepcopy(normalized_run_config)

    # Assumes partition key already validated
    def get_tags_for_partition_key(
        self,
        partition_key: str,
        job_name: str | None = None,
    ) -> Mapping[str, str]:
        from dagster._core.remote_representation.external_data import (
            partition_set_snap_name_for_job_name,
        )

        # _tags_for_partition_fn is deprecated, we can remove this branching logic in 2.0
        if self._tags_for_partition_fn:
            user_tags = self._tags_for_partition_fn(Partition(partition_key))
        elif self._tags_for_partition_key_fn:
            user_tags = self._tags_for_partition_key_fn(partition_key)
        else:
            user_tags = {}
        user_tags = normalize_tags(user_tags, allow_private_system_tags=False)

        system_tags = {
            **self.partitions_def.get_tags_for_partition_key(partition_key),
            # `PartitionSetDefinition` has been deleted but we still need to attach this special tag in
            # order for reexecution against partitions to work properly.
            **(
                {PARTITION_SET_TAG: partition_set_snap_name_for_job_name(job_name)}
                if job_name
                else {}
            ),
        }

        return {**user_tags, **system_tags}

    @classmethod
    def from_flexible_config(
        cls,
        config: Union[ConfigMapping, Mapping[str, object], "PartitionedConfig"] | None,
        partitions_def: PartitionsDefinition,
    ) -> "PartitionedConfig":
        check.invariant(
            not isinstance(config, ConfigMapping),
            "Can't supply a ConfigMapping for 'config' when 'partitions_def' is supplied.",
        )

        if isinstance(config, PartitionedConfig):
            check.invariant(
                config.partitions_def == partitions_def,
                "Can't supply a PartitionedConfig for 'config' with a different "
                "PartitionsDefinition than supplied for 'partitions_def'.",
            )
            return config
        else:
            hardcoded_config = config if config else {}
            return cls(
                partitions_def,  # type: ignore # ignored for update, fix me!
                run_config_for_partition_key_fn=lambda _: cast("Mapping", hardcoded_config),
            )

    def __call__(self, *args, **kwargs):
        if self._decorated_fn is None:
            raise DagsterInvalidInvocationError(
                "Only PartitionedConfig objects created using one of the partitioned config "
                "decorators can be directly invoked."
            )
        else:
            return self._decorated_fn(*args, **kwargs)


def partitioned_config(
    partitions_def: PartitionsDefinition,
    tags_for_partition_key_fn: Callable[[str], Mapping[str, str]] | None = None,
) -> Callable[[PartitionConfigFn], PartitionedConfig]:
    """Creates a partitioned config for a job given a PartitionsDefinition.

    The partitions_def provides the set of partitions, which may change over time
        (for example, when using a DynamicPartitionsDefinition).

    The decorated function takes in a partition key and returns a valid run config for a particular
    target job.

    Args:
        partitions_def: (Optional[DynamicPartitionsDefinition]): PartitionsDefinition for the job
        tags_for_partition_key_fn (Optional[Callable[[str], Mapping[str, str]]]): A function that
            accepts a partition key and returns a dictionary of tags to attach to runs for that
            partition.

    Returns:
        PartitionedConfig
    """
    check.opt_callable_param(tags_for_partition_key_fn, "tags_for_partition_key_fn")

    def inner(fn: PartitionConfigFn) -> PartitionedConfig:
        return PartitionedConfig(
            partitions_def=partitions_def,
            run_config_for_partition_key_fn=fn,
            decorated_fn=fn,
            tags_for_partition_key_fn=tags_for_partition_key_fn,
        )

    return inner



import functools
import hashlib
import heapq
import re
from collections.abc import Iterable, Iterator, Sequence
from datetime import date, datetime
from functools import cached_property
from typing import TYPE_CHECKING, Optional, cast

import dagster._check as check
from dagster._annotations import PublicAttr, public
from dagster._core.definitions.partitions.context import (
    PartitionLoadingContext,
    partition_loading_context,
)
from dagster._core.definitions.partitions.definition.partitions_definition import (
    PartitionsDefinition,
)
from dagster._core.definitions.partitions.partition_key_range import PartitionKeyRange
from dagster._core.definitions.partitions.schedule_type import (
    ScheduleType,
    cron_schedule_from_schedule_type_and_offsets,
)
from dagster._core.definitions.partitions.utils.time_window import TimeWindow, TimeWindowCursor
from dagster._core.definitions.timestamp import TimestampWithTimezone
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._core.types.pagination import PaginatedResults
from dagster._record import IHaveNew, record_custom
from dagster._serdes import whitelist_for_serdes
from dagster._time import (
    datetime_from_timestamp,
    dst_safe_strftime,
    dst_safe_strptime,
    get_timezone,
)
from dagster._utils.cronstring import get_fixed_minute_interval, is_basic_daily, is_basic_hourly
from dagster._utils.schedules import (
    cron_string_iterator,
    is_valid_cron_schedule,
    reverse_cron_string_iterator,
    schedule_execution_time_iterator,
)

if TYPE_CHECKING:
    from dagster._core.definitions.partitions.subset.partitions_subset import PartitionsSubset
    from dagster._core.definitions.partitions.subset.time_window import TimeWindowPartitionsSubset
    from dagster._core.instance import DynamicPartitionsStore


@whitelist_for_serdes
@record_custom(
    field_to_new_mapping={
        "start_ts": "start",
        "end_ts": "end",
    }
)
class TimeWindowPartitionsDefinition(PartitionsDefinition, IHaveNew):
    r"""A set of partitions where each partition corresponds to a time window.

    The provided cron_schedule determines the bounds of the time windows. E.g. a cron_schedule of
    "0 0 \\* \\* \\*" will result in daily partitions that start at midnight and end at midnight of the
    following day.

    The string partition_key associated with each partition corresponds to the start of the
    partition's time window.

    The first partition in the set will start on at the first cron_schedule tick that is equal to
    or after the given start datetime. The last partition in the set will end before the current
    time, unless the end_offset argument is set to a positive number.

    We recommended limiting partition counts for each asset to 100,000 partitions or fewer.

    Args:
        cron_schedule (str): Determines the bounds of the time windows.
        start (datetime): The first partition in the set will start on at the first cron_schedule
            tick that is equal to or after this value.
        timezone (Optional[str]): The timezone in which each time should exist.
            Supported strings for timezones are the ones provided by the
            `IANA time zone database <https://www.iana.org/time-zones>`_ - e.g. "America/Los_Angeles".

        end (datetime): The last partition (excluding) in the set.
        fmt (str): The date format to use for partition_keys. Note that if a non-UTC timezone is
            used, and the cron schedule repeats every hour or faster, the date format must include
            a timezone offset to disambiguate between multiple instances of the same time before and
            after the Fall DST transition. If the format does not contain this offset, the second
            instance of the ambiguous time partition key will have the UTC offset automatically
            appended to it.
        end_offset (int): Extends the partition set by a number of partitions equal to the value
            passed. If end_offset is 0 (the default), the last partition ends before the current
            time. If end_offset is 1, the second-to-last partition ends before the current time,
            and so on.
        exclusions (Optional[Sequence[Union[str, datetime]]]): Specifies a sequence of cron strings
            or datetime objects that should be excluded from the partition set. Every tick of the
            cron schedule that matches an excluded datetime or matches the tick of an excluded
            cron string will be excluded from the partition set.

    """

    start_ts: TimestampWithTimezone
    timezone: PublicAttr[str]
    end_ts: TimestampWithTimezone | None
    fmt: PublicAttr[str]
    end_offset: PublicAttr[int]
    cron_schedule: PublicAttr[str]
    exclusions: PublicAttr[Sequence[str | TimestampWithTimezone] | None]

    def __new__(
        cls,
        start: datetime | str | TimestampWithTimezone,
        fmt: str,
        end: datetime | str | TimestampWithTimezone | None = None,
        schedule_type: ScheduleType | None = None,
        timezone: str | None = None,
        end_offset: int = 0,
        minute_offset: int | None = None,
        hour_offset: int | None = None,
        day_offset: int | None = None,
        cron_schedule: str | None = None,
        exclusions: Sequence[str | datetime | TimestampWithTimezone] | None = None,
    ):
        check.opt_str_param(timezone, "timezone")
        timezone = timezone or "UTC"

        if isinstance(start, str):
            start_dt = dst_safe_strptime(start, timezone, fmt)
            start = TimestampWithTimezone(start_dt.timestamp(), timezone)
        elif isinstance(start, datetime):
            start_dt = start.replace(tzinfo=get_timezone(timezone))
            start = TimestampWithTimezone(start_dt.timestamp(), timezone)

        if not end:
            end = None
        elif isinstance(end, str):
            end_dt = dst_safe_strptime(end, timezone, fmt)
            end = TimestampWithTimezone(end_dt.timestamp(), timezone)
        elif isinstance(end, datetime):
            end_dt = end.replace(tzinfo=get_timezone(timezone))
            end = TimestampWithTimezone(end_dt.timestamp(), timezone)

        if cron_schedule is not None:
            check.invariant(
                schedule_type is None and not minute_offset and not hour_offset and not day_offset,
                "If cron_schedule argument is provided, then schedule_type, minute_offset, "
                "hour_offset, and day_offset can't also be provided",
            )
        else:
            if schedule_type is None:
                check.failed("One of schedule_type and cron_schedule must be provided")

            cron_schedule = cron_schedule_from_schedule_type_and_offsets(
                schedule_type=schedule_type,
                minute_offset=minute_offset or 0,
                hour_offset=hour_offset or 0,
                day_offset=day_offset or None,
            )

        if not is_valid_cron_schedule(cron_schedule):
            raise DagsterInvalidDefinitionError(
                f"Found invalid cron schedule '{cron_schedule}' for a"
                " TimeWindowPartitionsDefinition."
            )

        cleaned_exclusions: Sequence[str | TimestampWithTimezone] | None = None
        if exclusions:
            check.sequence_param(
                exclusions,
                "exclusions",
                of_type=(str, datetime, TimestampWithTimezone),
            )
            cron_exclusions = [cs for cs in exclusions if isinstance(cs, str)]
            invalid_exclusions = [cs for cs in cron_exclusions if not is_valid_cron_schedule(cs)]
            if invalid_exclusions:
                quoted_exclusions = [f"'{excl}'" for excl in invalid_exclusions]
                invalid_exclusion_str = ", ".join(quoted_exclusions)
                raise DagsterInvalidDefinitionError(
                    f"Found invalid cron schedule(s) {invalid_exclusion_str} in the exclusions"
                    " argument for a TimeWindowPartitionsDefinition. Expected a set of valid cron"
                    " strings and datetime objects."
                )
            cleaned_exclusions = []
            for exclusion_part in exclusions:
                if isinstance(exclusion_part, datetime):
                    dt = exclusion_part.replace(tzinfo=get_timezone(timezone))
                    cleaned_exclusions.append(TimestampWithTimezone(dt.timestamp(), timezone))
                else:
                    cleaned_exclusions.append(exclusion_part)

        return super().__new__(
            cls,
            start_ts=start,
            timezone=timezone,
            end_ts=end,
            fmt=fmt,
            end_offset=end_offset,
            cron_schedule=cron_schedule,
            exclusions=cleaned_exclusions if cleaned_exclusions else None,
        )

    def validate_partition_definition(self) -> None:
        # Try to determine if there are multiple time ranges being mapped
        # to the same partition key.
        first_partition_key = self.get_first_partition_key()

        if first_partition_key is None:
            return

        if self.get_next_partition_key(first_partition_key) == first_partition_key:
            raise DagsterInvalidDefinitionError(
                "This partition set contains multiple time ranges that map to the same partition key. "
                f"This usually indicates that the partition set's format string ({self.fmt}) is "
                f"not granular enough to produce a unique key for each time in the cron schedule ({self.cron_schedule})."
            )

    @property
    def start_timestamp(self) -> float:
        return self.start_ts.timestamp

    @property
    def end_timestamp(self) -> float | None:
        return self.end_ts.timestamp if self.end_ts else None

    @public
    @cached_property
    def start(self) -> datetime:
        start_timestamp_with_timezone = self.start_ts
        return datetime_from_timestamp(
            start_timestamp_with_timezone.timestamp, start_timestamp_with_timezone.timezone
        )

    @public
    @cached_property
    def end(self) -> datetime | None:
        end_timestamp_with_timezone = self.end_ts

        if not end_timestamp_with_timezone:
            return None

        return datetime.fromtimestamp(
            end_timestamp_with_timezone.timestamp,
            get_timezone(end_timestamp_with_timezone.timezone),
        )

    def _get_current_timestamp(self) -> float:
        with partition_loading_context() as ctx:
            current_time = ctx.effective_dt
            # if a naive current time was passed in, assume it was the same timezone as the partition set
            if not current_time.tzinfo:
                current_time = current_time.replace(tzinfo=get_timezone(self.timezone))

            return current_time.timestamp()

    def get_num_partitions_in_window(self, time_window: TimeWindow) -> int:
        if time_window.start.timestamp() >= time_window.end.timestamp():
            return 0
        if self.is_basic_daily:
            return (
                date(
                    time_window.end.year,
                    time_window.end.month,
                    time_window.end.day,
                )
                - date(
                    time_window.start.year,
                    time_window.start.month,
                    time_window.start.day,
                )
            ).days

        fixed_minute_interval = get_fixed_minute_interval(self.cron_schedule)
        if fixed_minute_interval and not self.exclusions:
            minutes_in_window = (time_window.end.timestamp() - time_window.start.timestamp()) / 60
            return int(minutes_in_window // fixed_minute_interval)

        return len(self.get_partition_keys_in_time_window(time_window))

    def get_num_partitions(self) -> int:
        last_partition_window = self.get_last_partition_window()
        first_partition_window = self.get_first_partition_window()

        if not last_partition_window or not first_partition_window:
            return 0

        return self.get_num_partitions_in_window(
            TimeWindow(start=first_partition_window.start, end=last_partition_window.end)
        )

    def get_partition_keys_between_indexes(self, start_idx: int, end_idx: int) -> list[str]:
        # Fetches the partition keys between the given start and end indices.
        # Start index is inclusive, end index is exclusive.
        # Method added for performance reasons, to only string format
        # partition keys included within the indices.
        current_timestamp = self._get_current_timestamp()

        partitions_past_current_time = 0
        partition_keys = []
        reached_end = False

        for idx, time_window in enumerate(self._iterate_time_windows(self.start_timestamp)):
            if time_window.end.timestamp() >= current_timestamp:
                reached_end = True
            if self.end_timestamp is not None and time_window.end.timestamp() > self.end_timestamp:
                reached_end = True
            if (
                time_window.end.timestamp() <= current_timestamp
                or partitions_past_current_time < self.end_offset
            ):
                if idx >= start_idx and idx < end_idx:
                    partition_keys.append(
                        dst_safe_strftime(
                            time_window.start, self.timezone, self.fmt, self.cron_schedule
                        )
                    )
                if time_window.end.timestamp() > current_timestamp:
                    partitions_past_current_time += 1
            else:
                break
            if len(partition_keys) >= end_idx - start_idx:
                break

        if reached_end and self.end_offset < 0:
            partition_keys = partition_keys[: self.end_offset]

        return partition_keys

    def is_window_start_excluded(self, window_start: datetime):
        if not self.exclusions:
            return False

        start_ts = TimestampWithTimezone(
            window_start.replace(tzinfo=get_timezone(self.timezone)).timestamp(), self.timezone
        )
        if start_ts in self.exclusions:
            return True

        excluded_cron_strings = [
            cron_string for cron_string in self.exclusions if isinstance(cron_string, str)
        ]
        for cron_string in excluded_cron_strings:
            if window_start == next(
                cron_string_iterator(window_start.timestamp(), cron_string, self.timezone)
            ):
                return True
        return False

    def get_partition_keys(
        self,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> Sequence[str]:
        with partition_loading_context(current_time, dynamic_partitions_store):
            current_timestamp = self._get_current_timestamp()

            last_partition_window = self._get_last_partition_window(current_timestamp)
            if not last_partition_window:
                return []

            partition_keys = []

            for time_window in self._iterate_time_windows(self.start_timestamp):
                partition_keys.append(
                    dst_safe_strftime(
                        time_window.start, self.timezone, self.fmt, self.cron_schedule
                    )
                )

                if time_window.start.timestamp() >= last_partition_window.start.timestamp():
                    break

            return partition_keys

    def get_paginated_partition_keys(
        self,
        context: PartitionLoadingContext,
        limit: int,
        ascending: bool,
        cursor: str | None = None,
    ) -> PaginatedResults[str]:
        with partition_loading_context(new_ctx=context):
            current_timestamp = self._get_current_timestamp()

            if cursor:
                time_window_cursor = TimeWindowCursor.from_cursor(cursor)
                start_timestamp = time_window_cursor.start_timestamp
                end_timestamp = time_window_cursor.end_timestamp
                offset_partitions_count = time_window_cursor.offset_partition_count
            else:
                start_timestamp = self.start_timestamp
                end_timestamp = current_timestamp
                offset_partitions_count = 0

            partition_keys: list[str] = []
            has_more = False

            if not ascending:
                offset_time_windows = []
                if self.end_offset > 0 and self.end_offset > offset_partitions_count:
                    last_no_offset_time_window = next(
                        iter(self._reverse_iterate_time_windows(end_timestamp)), None
                    )
                    lookforward_time = (
                        last_no_offset_time_window.end.timestamp()
                        if last_no_offset_time_window
                        else end_timestamp
                    )
                    for time_window in self._iterate_time_windows(lookforward_time):
                        if (
                            self.end_timestamp is not None
                            and time_window.end.timestamp() > self.end_timestamp
                        ):
                            break

                        if len(offset_time_windows) >= self.end_offset - offset_partitions_count:
                            break

                        if not self.is_window_start_excluded(time_window.start):
                            offset_time_windows.append(
                                dst_safe_strftime(
                                    time_window.start, self.timezone, self.fmt, self.cron_schedule
                                )
                            )

                partition_keys = list(reversed(offset_time_windows))[:limit]
                offset_partitions_count += len(partition_keys)

                for time_window in self._reverse_iterate_time_windows(end_timestamp):
                    if len(partition_keys) >= limit - min(0, self.end_offset):
                        has_more = True
                        break

                    if time_window.start.timestamp() < start_timestamp:
                        break

                    partition_keys.append(
                        dst_safe_strftime(
                            time_window.start, self.timezone, self.fmt, self.cron_schedule
                        )
                    )

                if self.end_offset < 0 and not cursor:
                    # only subset if we did not have a cursor... if we did have a cursor, we've already
                    # applied the offset to the end, since we're moving backwards from the end
                    partition_keys = partition_keys[-1 * min(0, self.end_offset) :]

            else:
                for time_window in self._iterate_time_windows(start_timestamp):
                    if (
                        self.end_timestamp is not None
                        and time_window.end.timestamp() > self.end_timestamp
                    ):
                        break
                    if (
                        time_window.end.timestamp() <= end_timestamp
                        or offset_partitions_count < self.end_offset
                    ):
                        partition_keys.append(
                            dst_safe_strftime(
                                time_window.start, self.timezone, self.fmt, self.cron_schedule
                            )
                        )
                        if time_window.end.timestamp() > end_timestamp:
                            offset_partitions_count += 1
                        if len(partition_keys) >= limit - min(0, self.end_offset):
                            has_more = True
                            break
                    else:
                        break

                if has_more:
                    # exited due to limit; subset in case we overshot (if end_offset < 0)
                    partition_keys = partition_keys[:limit]
                elif self.end_offset < 0:
                    # only subset if we did not eject early due to the limit
                    partition_keys = partition_keys[: self.end_offset]

            if not partition_keys:
                next_cursor = TimeWindowCursor(
                    start_timestamp=int(start_timestamp),
                    end_timestamp=int(end_timestamp),
                    offset_partition_count=offset_partitions_count,
                )
            elif ascending:
                last_partition_key = partition_keys[-1]
                last_time_window = self.time_window_for_partition_key(last_partition_key)
                next_cursor = TimeWindowCursor(
                    start_timestamp=int(last_time_window.end.timestamp()),
                    end_timestamp=int(end_timestamp),
                    offset_partition_count=offset_partitions_count,
                )
            else:
                last_partition_key = partition_keys[-1]
                last_time_window = self.time_window_for_partition_key(last_partition_key)
                next_cursor = TimeWindowCursor(
                    start_timestamp=int(start_timestamp),
                    end_timestamp=int(end_timestamp)
                    if self.end_offset > 0 and offset_partitions_count < self.end_offset
                    else int(last_time_window.start.timestamp()),
                    offset_partition_count=offset_partitions_count,
                )

            return PaginatedResults(
                results=partition_keys,
                cursor=str(next_cursor),
                has_more=has_more,
            )

    def __str__(self) -> str:
        schedule_str = (
            self.schedule_type.value.capitalize() if self.schedule_type else self.cron_schedule
        )
        partition_def_str = f"{schedule_str}, starting {dst_safe_strftime(self.start, self.timezone, self.fmt, self.cron_schedule)} {self.timezone}."
        if self.end_offset != 0:
            partition_def_str += (
                " End offsetted by"
                f" {self.end_offset} partition{'' if self.end_offset == 1 else 's'}."
            )
        return partition_def_str

    def __repr__(self):
        # Between python 3.8 and 3.9 the repr of a datetime object changed.
        # Replaces start time with timestamp as a workaround to make sure the repr is consistent across versions.
        # Make sure to update this __repr__ if any new fields are added to TimeWindowPartitionsDefinition.
        exclusions_str = f", exclusions={self.exclusions}" if self.exclusions else ""
        return (
            f"TimeWindowPartitionsDefinition(start={self.start_timestamp},"
            f" end={self.end_timestamp if self.end_timestamp is not None else None},"
            f" timezone='{self.timezone}', fmt='{self.fmt}', end_offset={self.end_offset},"
            f" cron_schedule='{self.cron_schedule}')"
        ) + exclusions_str

    def __hash__(self):
        return hash(tuple(self.__repr__()))

    @functools.lru_cache(maxsize=100)
    def time_window_for_partition_key(self, partition_key: str) -> TimeWindow:
        partition_key_dt = dst_safe_strptime(partition_key, self.timezone, self.fmt)
        return next(iter(self._iterate_time_windows(partition_key_dt.timestamp())))

    @functools.lru_cache(maxsize=5)
    def time_windows_for_partition_keys(
        self, partition_keys: frozenset[str], validate: bool = True
    ) -> Sequence[TimeWindow]:
        if len(partition_keys) == 0:
            return []

        sorted_pks = sorted(
            partition_keys,
            key=lambda pk: dst_safe_strptime(pk, self.timezone, self.fmt).timestamp(),
        )
        cur_windows_iterator = iter(
            self._iterate_time_windows(
                dst_safe_strptime(sorted_pks[0], self.timezone, self.fmt).timestamp()
            )
        )
        partition_key_time_windows: list[TimeWindow] = []
        for partition_key in sorted_pks:
            next_window = next(cur_windows_iterator)
            if (
                dst_safe_strftime(next_window.start, self.timezone, self.fmt, self.cron_schedule)
                == partition_key
            ):
                partition_key_time_windows.append(next_window)
            else:
                cur_windows_iterator = iter(
                    self._iterate_time_windows(
                        dst_safe_strptime(partition_key, self.timezone, self.fmt).timestamp()
                    )
                )
                partition_key_time_windows.append(next(cur_windows_iterator))

        if validate:
            start_time_window = self.get_first_partition_window()
            end_time_window = self.get_last_partition_window()

            if start_time_window is None or end_time_window is None:
                check.failed("No partitions in the PartitionsDefinition")

            start_timestamp = start_time_window.start.timestamp()
            end_timestamp = end_time_window.end.timestamp()

            partition_key_time_windows = [
                tw
                for tw in partition_key_time_windows
                if tw.start.timestamp() >= start_timestamp and tw.end.timestamp() <= end_timestamp
            ]
        return partition_key_time_windows

    def start_time_for_partition_key(self, partition_key: str) -> datetime:
        partition_key_dt = dst_safe_strptime(partition_key, self.timezone, self.fmt)
        if self.is_basic_hourly or self.is_basic_daily:
            return partition_key_dt
        # the datetime format might not include granular components, so we need to recover them,
        # e.g. if cron_schedule="0 7 * * *" and fmt="%Y-%m-%d".
        # we make the assumption that the parsed partition key is <= the start datetime.
        return next(iter(self._iterate_time_windows(partition_key_dt.timestamp()))).start

    def get_next_partition_key(self, partition_key: str) -> str | None:
        last_partition_window = self.get_last_partition_window()
        if last_partition_window is None:
            return None

        partition_key_dt = dst_safe_strptime(partition_key, self.timezone, self.fmt)
        windows_iter = iter(self._iterate_time_windows(partition_key_dt.timestamp()))
        next(windows_iter)
        start_time = next(windows_iter).start
        if start_time.timestamp() >= last_partition_window.end.timestamp():
            return None
        else:
            return dst_safe_strftime(start_time, self.timezone, self.fmt, self.cron_schedule)

    def get_next_partition_window(
        self, end_dt: datetime, respect_bounds: bool = True
    ) -> TimeWindow | None:
        windows_iter = iter(self._iterate_time_windows(end_dt.timestamp()))
        next_window = next(windows_iter)

        if respect_bounds:
            last_partition_window = self.get_last_partition_window()
            if last_partition_window is None:
                return None

            if next_window.start.timestamp() >= last_partition_window.end.timestamp():
                return None

        return next_window

    def get_prev_partition_window(
        self, start_dt: datetime, respect_bounds: bool = True
    ) -> TimeWindow | None:
        windows_iter = iter(self._reverse_iterate_time_windows(start_dt.timestamp()))
        prev_window = next(windows_iter)
        if respect_bounds:
            first_partition_window = self.get_first_partition_window()
            if (
                first_partition_window is None
                or prev_window.start.timestamp() < first_partition_window.start.timestamp()
            ):
                return None

        return prev_window

    @functools.lru_cache(maxsize=256)
    def _get_first_partition_window(self, current_timestamp: float) -> TimeWindow | None:
        time_window = next(iter(self._iterate_time_windows(self.start_timestamp)))

        if self.end_timestamp is not None and self.end_timestamp <= self.start_timestamp:
            return None

        if self.end_offset == 0:
            return time_window if time_window.end.timestamp() <= current_timestamp else None
        elif self.end_offset > 0:
            iterator = iter(self._iterate_time_windows(current_timestamp))
            # first returned time window is time window of current time
            curr_window_plus_offset = next(iterator)
            for _ in range(self.end_offset):
                curr_window_plus_offset = next(iterator)
            return (
                time_window
                if time_window.end.timestamp() <= curr_window_plus_offset.start.timestamp()
                else None
            )
        else:
            # end offset < 0
            end_window = None
            iterator = iter(self._reverse_iterate_time_windows(current_timestamp))
            for _ in range(abs(self.end_offset)):
                end_window = next(iterator)

            if end_window is None:
                check.failed("end_window should not be None")

            return (
                time_window if time_window.end.timestamp() <= end_window.start.timestamp() else None
            )

    def get_first_partition_window(self) -> TimeWindow | None:
        return self._get_first_partition_window(self._get_current_timestamp())

    @functools.lru_cache(maxsize=256)
    def _get_last_partition_window(
        self, current_timestamp: float, ignore_exclusions: bool = False
    ) -> TimeWindow | None:
        first_window = self.get_first_partition_window()
        if first_window is None:
            return None

        if self.end_offset == 0:
            if self.end_timestamp is not None and self.end_timestamp < current_timestamp:
                current_timestamp = self.end_timestamp

            return next(
                iter(
                    self._reverse_iterate_time_windows(
                        current_timestamp, ignore_exclusions=ignore_exclusions
                    )
                )
            )

        last_window_before_end_timestamp = None
        current_timestamp_window = None

        if self.end_timestamp is not None:
            last_window_before_end_timestamp = next(
                iter(
                    self._reverse_iterate_time_windows(
                        self.end_timestamp, ignore_exclusions=ignore_exclusions
                    )
                )
            )

        current_timestamp_iter = iter(
            self._reverse_iterate_time_windows(
                current_timestamp, ignore_exclusions=ignore_exclusions
            )
        )
        # first returned time window is the last window <= the current timestamp
        end_offset_zero_window = next(current_timestamp_iter)

        if self.end_offset < 0:
            for _ in range(abs(self.end_offset)):
                current_timestamp_window = next(current_timestamp_iter)
        else:
            current_timestamp_iter = iter(
                self._iterate_time_windows(
                    end_offset_zero_window.end.timestamp(), ignore_exclusions=ignore_exclusions
                )
            )
            for _ in range(self.end_offset):
                current_timestamp_window = next(current_timestamp_iter)

        current_timestamp_window = check.not_none(
            current_timestamp_window,
            "current_timestamp_window should not be None if end_offset != 0",
        )

        if (
            last_window_before_end_timestamp
            and last_window_before_end_timestamp.start.timestamp()
            <= current_timestamp_window.start.timestamp()
        ):
            return last_window_before_end_timestamp
        elif current_timestamp_window.start.timestamp() < first_window.start.timestamp():
            return first_window
        else:
            return current_timestamp_window

    def get_last_partition_window(self) -> TimeWindow | None:
        return self._get_last_partition_window(
            self._get_current_timestamp(), ignore_exclusions=False
        )

    def get_last_partition_window_ignoring_exclusions(self) -> TimeWindow | None:
        return self._get_last_partition_window(
            self._get_current_timestamp(), ignore_exclusions=True
        )

    def get_first_partition_key(self) -> str | None:
        first_window = self.get_first_partition_window()
        if first_window is None:
            return None

        return dst_safe_strftime(first_window.start, self.timezone, self.fmt, self.cron_schedule)

    def get_last_partition_key(self) -> str | None:
        last_window = self.get_last_partition_window()
        if last_window is None:
            return None

        return dst_safe_strftime(last_window.start, self.timezone, self.fmt, self.cron_schedule)

    def end_time_for_partition_key(self, partition_key: str) -> datetime:
        return self.time_window_for_partition_key(partition_key).end

    @functools.lru_cache(maxsize=5)
    def get_partition_keys_in_time_window(self, time_window: TimeWindow) -> Sequence[str]:
        result: list[str] = []
        time_window_end_timestamp = time_window.end.timestamp()
        for partition_time_window in self._iterate_time_windows(time_window.start.timestamp()):
            if partition_time_window.start.timestamp() < time_window_end_timestamp:
                result.append(
                    dst_safe_strftime(
                        partition_time_window.start, self.timezone, self.fmt, self.cron_schedule
                    )
                )
            else:
                break
        return result

    def get_partition_subset_in_time_window(
        self, time_window: TimeWindow
    ) -> "TimeWindowPartitionsSubset":
        return self.partitions_subset_class(
            partitions_def=self, num_partitions=None, included_time_windows=[time_window]
        )

    def get_partition_key_range_for_time_window(
        self, time_window: TimeWindow, respect_bounds: bool = True
    ) -> PartitionKeyRange:
        start_partition_key = self.get_partition_key_for_timestamp(time_window.start.timestamp())
        end_partition_key = self.get_partition_key_for_timestamp(
            check.not_none(
                self.get_prev_partition_window(time_window.end, respect_bounds=respect_bounds)
            ).start.timestamp()
        )

        return PartitionKeyRange(start_partition_key, end_partition_key)

    def get_partition_keys_in_range(self, partition_key_range: PartitionKeyRange) -> Sequence[str]:
        start_time = self.start_time_for_partition_key(partition_key_range.start)
        check.invariant(
            start_time.timestamp() >= self.start_timestamp,
            (
                f"Partition key range start {partition_key_range.start} is before "
                f"the partitions definition start time {self.start}"
            ),
        )
        end_time = self.end_time_for_partition_key(partition_key_range.end)
        if self.end_timestamp is not None:
            check.invariant(
                end_time.timestamp() <= self.end_timestamp,
                (
                    f"Partition key range end {partition_key_range.end} is after the "
                    f"partitions definition end time {self.end}"
                ),
            )

        return self.get_partition_keys_in_time_window(TimeWindow(start_time, end_time))

    @public
    @property
    def schedule_type(self) -> ScheduleType | None:
        """Optional[ScheduleType]: An enum representing the partition cadence (hourly, daily,
        weekly, or monthly).
        """
        if re.fullmatch(r"\d+ \* \* \* \*", self.cron_schedule):
            return ScheduleType.HOURLY
        elif re.fullmatch(r"\d+ \d+ \* \* \*", self.cron_schedule):
            return ScheduleType.DAILY
        elif re.fullmatch(r"\d+ \d+ \* \* \d+", self.cron_schedule):
            return ScheduleType.WEEKLY
        elif re.fullmatch(r"\d+ \d+ \d+ \* \*", self.cron_schedule):
            return ScheduleType.MONTHLY
        else:
            return None

    @public
    @property
    def minute_offset(self) -> int:
        """int: Number of minutes past the hour to "split" partitions. Defaults to 0.

        For example, returns 15 if each partition starts at 15 minutes past the hour.
        """
        match = re.fullmatch(r"(\d+) (\d+|\*) (\d+|\*) (\d+|\*) (\d+|\*)", self.cron_schedule)
        if match is None:
            check.failed(f"{self.cron_schedule} has no minute offset")
        return int(match.groups()[0])

    @public
    @property
    def hour_offset(self) -> int:
        """int: Number of hours past 00:00 to "split" partitions. Defaults to 0.

        For example, returns 1 if each partition starts at 01:00.
        """
        match = re.fullmatch(r"(\d+|\*) (\d+) (\d+|\*) (\d+|\*) (\d+|\*)", self.cron_schedule)
        if match is None:
            check.failed(f"{self.cron_schedule} has no hour offset")
        return int(match.groups()[1])

    @public
    @property
    def day_offset(self) -> int:
        """int: For a weekly or monthly partitions definition, returns the day to "split" partitions
        by. Each partition will start on this day, and end before this day in the following
        week/month. Returns 0 if the day_offset parameter is unset in the
        WeeklyPartitionsDefinition, MonthlyPartitionsDefinition, or the provided cron schedule.

        For weekly partitions, returns a value between 0 (representing Sunday) and 6 (representing
        Saturday). Providing a value of 1 means that a partition will exist weekly from Monday to
        the following Sunday.

        For monthly partitions, returns a value between 0 (the first day of the month) and 31 (the
        last possible day of the month).
        """
        schedule_type = self.schedule_type
        if schedule_type == ScheduleType.WEEKLY:
            match = re.fullmatch(r"(\d+|\*) (\d+|\*) (\d+|\*) (\d+|\*) (\d+)", self.cron_schedule)
            if match is None:
                check.failed(f"{self.cron_schedule} has no day offset")
            return int(match.groups()[4])
        elif schedule_type == ScheduleType.MONTHLY:
            match = re.fullmatch(r"(\d+|\*) (\d+|\*) (\d+) (\d+|\*) (\d+|\*)", self.cron_schedule)
            if match is None:
                check.failed(f"{self.cron_schedule} has no day offset")
            return int(match.groups()[2])
        else:
            check.failed(f"Unsupported schedule type for day_offset: {schedule_type}")

    @public
    def get_cron_schedule(
        self,
        minute_of_hour: int | None = None,
        hour_of_day: int | None = None,
        day_of_week: int | None = None,
        day_of_month: int | None = None,
    ) -> str:
        """The schedule executes at the cadence specified by the partitioning, but may overwrite
        the minute/hour/day offset of the partitioning.

        This is useful e.g. if you have partitions that span midnight to midnight but you want to
        schedule a job that runs at 2 am.
        """
        if (
            minute_of_hour is None
            and hour_of_day is None
            and day_of_week is None
            and day_of_month is None
        ):
            return self.cron_schedule

        schedule_type = self.schedule_type
        if schedule_type is None:
            check.failed(
                f"{self.cron_schedule} does not support"
                " minute_of_hour/hour_of_day/day_of_week/day_of_month arguments"
            )

        minute_of_hour = cast(
            "int",
            check.opt_int_param(minute_of_hour, "minute_of_hour", default=self.minute_offset),
        )

        if schedule_type == ScheduleType.HOURLY:
            check.invariant(
                hour_of_day is None, "Cannot set hour parameter with hourly partitions."
            )
        else:
            hour_of_day = cast(
                "int", check.opt_int_param(hour_of_day, "hour_of_day", default=self.hour_offset)
            )

        if schedule_type == ScheduleType.DAILY:
            check.invariant(
                day_of_week is None, "Cannot set day of week parameter with daily partitions."
            )
            check.invariant(
                day_of_month is None, "Cannot set day of month parameter with daily partitions."
            )

        if schedule_type == ScheduleType.MONTHLY:
            default = self.day_offset or 1
            day_offset = check.opt_int_param(day_of_month, "day_of_month", default=default)
        elif schedule_type == ScheduleType.WEEKLY:
            default = self.day_offset or 0
            day_offset = check.opt_int_param(day_of_week, "day_of_week", default=default)
        else:
            day_offset = 0

        return cron_schedule_from_schedule_type_and_offsets(
            schedule_type,
            minute_offset=minute_of_hour,
            hour_offset=hour_of_day or 0,
            day_offset=day_offset,
        )

    def _build_exclusion_iterator_list(
        self, boundary_timestamp: float, ascending: bool
    ) -> list[Iterator[datetime]]:
        if not self.exclusions:
            return []

        # Create a single iterator for filtered and sorted timestamps
        def _timestamp_iter():
            # Collect and sort timestamps, then filter based on boundary
            timestamps = [
                excl.timestamp
                for excl in self.exclusions  # type: ignore
                if isinstance(excl, TimestampWithTimezone)
            ]
            # Sort based on direction: ascending or descending
            timestamps.sort(reverse=not ascending)

            for ts in timestamps:
                # Filter based on boundary
                if (ascending and ts < boundary_timestamp) or (
                    not ascending and ts > boundary_timestamp
                ):
                    continue
                yield datetime_from_timestamp(ts, self.timezone)

        excluded_cron_strings = [excl for excl in self.exclusions if isinstance(excl, str)]
        excluded_cron_strings_iters = (
            [
                schedule_execution_time_iterator(
                    start_timestamp=boundary_timestamp,
                    cron_schedule=excluded_cron_strings,
                    execution_timezone=self.timezone,
                    ascending=ascending,
                )
            ]
            if excluded_cron_strings
            else []
        )

        return [_timestamp_iter(), *excluded_cron_strings_iters]

    def _exclusion_iterator(self, start_timestamp: float) -> Iterator[datetime]:
        exclusion_iterators = self._build_exclusion_iterator_list(start_timestamp, ascending=True)
        yield from heapq.merge(
            *[excl_iter for excl_iter in exclusion_iterators], key=lambda x: x.timestamp()
        )

    def _reverse_exclusion_iterator(self, end_timestamp: float) -> Iterator[datetime]:
        exclusion_iterators = self._build_exclusion_iterator_list(end_timestamp, ascending=False)
        yield from heapq.merge(
            *[excl_iter for excl_iter in exclusion_iterators], key=lambda x: -x.timestamp()
        )

    def _iterate_time_windows(
        self, start_timestamp: float, ignore_exclusions: bool = False
    ) -> Iterable[TimeWindow]:
        """Returns an infinite generator of time windows that start >= the given start time."""
        iterator = cron_string_iterator(
            start_timestamp=start_timestamp,
            cron_string=self.cron_schedule,
            execution_timezone=self.timezone,
        )
        curr_time = next(iterator)
        while curr_time.timestamp() < start_timestamp:
            curr_time = next(iterator)

        exclusion_iter = None
        next_excluded_time = None
        if not ignore_exclusions:
            exclusion_iter = self._exclusion_iterator(start_timestamp)
            next_excluded_time = next(exclusion_iter, None)

        while True:
            next_time = next(iterator)

            is_excluded = False
            if exclusion_iter and next_excluded_time:
                curr_timestamp = curr_time.timestamp()

                while (
                    next_excluded_time is not None
                    and next_excluded_time.timestamp() < curr_timestamp
                ):
                    next_excluded_time = next(exclusion_iter, None)

                if (
                    next_excluded_time is not None
                    and next_excluded_time.timestamp() == curr_timestamp
                ):
                    is_excluded = True

            if not is_excluded:
                yield TimeWindow(curr_time, next_time)

            curr_time = next_time

    def _reverse_iterate_time_windows(
        self, end_timestamp: float, ignore_exclusions: bool = False
    ) -> Iterable[TimeWindow]:
        """Returns an infinite generator of time windows that end before the given end timestamp.
        For example, if you pass in any time on day N (including midnight) for a daily partition
        with offset 0 bounded at midnight, the first element this iterator will return is
        [day N-1, day N).

        If ignore_exclusions is True, excluded windows will be included in the iteration.  This is
        useful for checking for skipping excluded windows when calculating a schedule off of the
        time window partitions definition
        """
        iterator = reverse_cron_string_iterator(
            end_timestamp=end_timestamp,
            cron_string=self.cron_schedule,
            execution_timezone=self.timezone,
        )

        curr_time = next(iterator)
        while curr_time.timestamp() > end_timestamp:
            curr_time = next(iterator)

        exclusion_iter = None
        next_excluded_time = None
        if not ignore_exclusions:
            exclusion_iter = self._reverse_exclusion_iterator(end_timestamp)
            next_excluded_time = next(exclusion_iter, None)

        while True:
            prev_time = next(iterator)

            is_excluded = False
            if exclusion_iter and next_excluded_time is not None:
                prev_timestamp = prev_time.timestamp()

                while (
                    next_excluded_time is not None
                    and next_excluded_time.timestamp() > prev_timestamp
                ):
                    next_excluded_time = next(exclusion_iter, None)

                if (
                    next_excluded_time is not None
                    and next_excluded_time.timestamp() == prev_timestamp
                ):
                    is_excluded = True

            if not is_excluded:
                yield TimeWindow(prev_time, curr_time)

            curr_time = prev_time

    def get_partition_key_for_timestamp(self, timestamp: float, end_closed: bool = False) -> str:
        """Args:
        timestamp (float): Timestamp from the unix epoch, UTC.
        end_closed (bool): Whether the interval is closed at the end or at the beginning.
        """
        rev_iter = reverse_cron_string_iterator(timestamp, self.cron_schedule, self.timezone)
        prev_partition_key = None
        while prev_partition_key is None:
            prev_dt = next(rev_iter)
            if end_closed and prev_dt.timestamp() == timestamp:
                continue

            if self.is_window_start_excluded(prev_dt):
                continue

            prev_partition_key = dst_safe_strftime(
                prev_dt, self.timezone, self.fmt, self.cron_schedule
            )

        iterator = cron_string_iterator(timestamp, self.cron_schedule, self.timezone)
        next_partition_key = None
        next_dt = None
        while next_partition_key is None:
            next_dt = next(iterator)
            if self.is_window_start_excluded(next_dt):
                continue
            next_partition_key = dst_safe_strftime(
                next_dt, self.timezone, self.fmt, self.cron_schedule
            )

        if end_closed or (next_dt and next_dt.timestamp() > timestamp):
            return prev_partition_key
        else:
            return next_partition_key

    def less_than(self, partition_key1: str, partition_key2: str) -> bool:
        """Returns true if the partition_key1 is earlier than partition_key2."""
        return (
            self.start_time_for_partition_key(partition_key1).timestamp()
            < self.start_time_for_partition_key(partition_key2).timestamp()
        )

    @property
    def partitions_subset_class(self) -> type["TimeWindowPartitionsSubset"]:
        from dagster._core.definitions.partitions.subset.time_window import (
            TimeWindowPartitionsSubset,
        )

        return TimeWindowPartitionsSubset

    def subset_with_all_partitions(self) -> "PartitionsSubset":
        first_window = self.get_first_partition_window()
        last_window = self.get_last_partition_window()
        windows = (
            []
            if first_window is None or last_window is None
            else [TimeWindow(first_window.start, last_window.end)]
        )
        return self.partitions_subset_class(
            partitions_def=self, num_partitions=None, included_time_windows=windows
        )

    def get_serializable_unique_identifier(
        self, dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None
    ) -> str:
        return hashlib.sha1(self.__repr__().encode("utf-8")).hexdigest()

    def has_partition_key(self, partition_key: str) -> bool:
        """Returns a boolean representing if the given partition key is valid."""
        try:
            partition_start_time = self.start_time_for_partition_key(partition_key)
            partition_start_timestamp = partition_start_time.timestamp()
        except ValueError:
            # unparseable partition key
            return False

        if self.is_window_start_excluded(partition_start_time):
            return False

        first_partition_window = self.get_first_partition_window()
        last_partition_window = self.get_last_partition_window()
        return not (
            # no partitions at all
            first_partition_window is None
            or last_partition_window is None
            # partition starts before the first valid partition
            or partition_start_timestamp < first_partition_window.start.timestamp()
            # partition starts after the last valid partition
            or partition_start_timestamp > last_partition_window.start.timestamp()
            # partition key string does not represent the start of an actual partition
            or dst_safe_strftime(partition_start_time, self.timezone, self.fmt, self.cron_schedule)
            != partition_key
        )

    def equal_except_for_start_or_end(self, other: "TimeWindowPartitionsDefinition") -> bool:
        """Returns True iff this is identical to other, except they're allowed to have different
        start and end datetimes.
        """
        return (
            self.timezone == other.timezone
            and self.fmt == other.fmt
            and self.cron_schedule == other.cron_schedule
            and self.end_offset == other.end_offset
            and self.exclusions == other.exclusions
        )

    def get_partition_key(self, key: str | date | datetime) -> str:
        if isinstance(key, date) or isinstance(key, datetime):
            key = key.strftime(self.fmt)

        # now should have str
        check.str_param(key, "key")
        if not self.has_partition_key(key):
            raise ValueError(f"Got invalid partition key {key!r}")

        return key

    @property
    def is_basic_daily(self) -> bool:
        return not self.exclusions and is_basic_daily(self.cron_schedule)

    @property
    def is_basic_hourly(self) -> bool:
        return not self.exclusions and is_basic_hourly(self.cron_schedule)



import inspect
from collections.abc import AsyncIterator, Iterator, Mapping, Sequence
from typing import Any, TypeAlias, TypeVar

import dagster._check as check
from dagster._core.definitions import (
    AssetCheckEvaluation,
    AssetCheckResult,
    AssetMaterialization,
    AssetObservation,
    DynamicOutput,
    ExpectationResult,
    Node,
    NodeHandle,
    Output,
)
from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckKey
from dagster._core.definitions.assets.job.asset_layer import AssetLayer
from dagster._core.definitions.op_definition import OpComputeFunction
from dagster._core.definitions.result import AssetResult, MaterializeResult, ObserveResult
from dagster._core.errors import DagsterExecutionStepExecutionError, DagsterInvariantViolationError
from dagster._core.events import DagsterEvent
from dagster._core.execution.context.compute import ExecutionContextTypes
from dagster._core.execution.context.system import StepExecutionContext
from dagster._core.execution.plan.outputs import StepOutput, StepOutputProperties
from dagster._core.execution.plan.utils import op_execution_error_boundary
from dagster._core.system_config.objects import ResolvedRunConfig
from dagster._utils import iterate_with_context

T = TypeVar("T")

OpOutputUnion: TypeAlias = (
    DynamicOutput[Any]
    | Output[Any]
    | AssetMaterialization
    | ExpectationResult
    | AssetObservation
    | DagsterEvent
    | AssetCheckEvaluation
    | AssetCheckResult
    | MaterializeResult
    | ObserveResult
)


def create_step_outputs(
    node: Node,
    handle: NodeHandle,
    resolved_run_config: ResolvedRunConfig,
    asset_layer: AssetLayer,
) -> Sequence[StepOutput]:
    check.inst_param(node, "node", Node)
    check.inst_param(handle, "handle", NodeHandle)

    # the run config has the node output name configured
    config_output_names: set[str] = set()
    current_handle = handle
    while current_handle:
        op_config = resolved_run_config.ops[str(current_handle)]
        current_handle = current_handle.parent
        config_output_names = config_output_names.union(op_config.outputs.output_names)

    step_outputs: list[StepOutput] = []
    for name, output_def in node.definition.output_dict.items():
        asset_key = asset_layer.get_asset_key_for_node_output(handle, name)
        asset_check_key = asset_layer.get_asset_check_key_for_node_output(handle, name)

        selected_entity_keys = asset_layer.get_selected_entity_keys_for_node(handle)
        asset_node = asset_layer.asset_graph.get(asset_key) if asset_key else None

        step_outputs.append(
            StepOutput(
                node_handle=handle,
                name=output_def.name,
                dagster_type_key=output_def.dagster_type.key,
                properties=StepOutputProperties(
                    is_required=output_def.is_required,
                    is_dynamic=output_def.is_dynamic,
                    is_asset=asset_key is not None,
                    should_materialize_DEPRECATED=output_def.name in config_output_names,
                    asset_key=asset_key if asset_key in selected_entity_keys else None,
                    is_asset_partitioned=bool(asset_node.partitions_def) if asset_node else False,
                    asset_check_key=asset_check_key
                    if asset_check_key in selected_entity_keys
                    else None,
                    asset_execution_type=asset_node.execution_type if asset_node else None,
                ),
            )
        )
    return step_outputs


def _validate_event(event: Any, step_context: StepExecutionContext) -> OpOutputUnion:
    if not isinstance(
        event,
        (
            DynamicOutput,
            Output,
            AssetMaterialization,
            ExpectationResult,
            AssetObservation,
            DagsterEvent,
            AssetCheckResult,
            AssetCheckEvaluation,
            MaterializeResult,
            ObserveResult,
        ),
    ):
        raise DagsterInvariantViolationError(
            f"Compute function for {step_context.describe_op()} yielded a value of type {type(event)} "
            "rather than an instance of Output, AssetMaterialization, or ExpectationResult."
            f" Values yielded by {step_context.op_def.node_type_str}s must be wrapped in one of these types. If your "
            f"{step_context.op_def.node_type_str} has a single output and yields no other events, you may want to use "
            f"`return` instead of `yield` in the body of your {step_context.op_def.node_type_str} compute function. If "
            "you are already using `return`, and you expected to return a value of type "
            f"{type(event)}, you may be inadvertently returning a generator rather than the value "
            # f"you expected. Value is {str(event[0])}"
        )

    return event


def gen_from_async_gen(
    context: StepExecutionContext,
    async_gen: AsyncIterator[T],
) -> Iterator[T]:
    while True:
        try:
            yield context.event_loop.run_until_complete(async_gen.__anext__())
        except StopAsyncIteration:
            return


def _yield_compute_results(
    step_context: StepExecutionContext,
    inputs: Mapping[str, Any],
    compute_fn: OpComputeFunction,
    compute_context: ExecutionContextTypes,
) -> Iterator[OpOutputUnion]:
    user_event_generator = compute_fn(compute_context, inputs)

    if isinstance(user_event_generator, Output):
        raise DagsterInvariantViolationError(
            f"Compute function for {step_context.describe_op()} returned an Output rather than "
            f"yielding it. The compute_fn of the {step_context.op_def.node_type_str} must yield "
            "its results"
        )

    if user_event_generator is None:
        return

    if inspect.isasyncgen(user_event_generator):
        user_event_generator = gen_from_async_gen(step_context, user_event_generator)

    op_label = step_context.describe_op()

    for event in iterate_with_context(
        lambda: op_execution_error_boundary(
            DagsterExecutionStepExecutionError,
            msg_fn=lambda: f"Error occurred while executing {op_label}:",
            step_context=step_context,
            step_key=step_context.step.key,
            op_def_name=step_context.op_def.name,
            op_name=step_context.op.name,
        ),
        user_event_generator,
    ):
        if compute_context.op_execution_context.has_events():
            yield from compute_context.op_execution_context.consume_events()
        yield _validate_event(event, step_context)

    if compute_context.op_execution_context.has_events():
        yield from compute_context.op_execution_context.consume_events()


def execute_core_compute(
    step_context: StepExecutionContext,
    inputs: Mapping[str, Any],
    compute_fn: OpComputeFunction,
    compute_context: ExecutionContextTypes,
) -> Iterator[OpOutputUnion]:
    """Execute the user-specified compute for the op. Wrap in an error boundary and do
    all relevant logging and metrics tracking.
    """
    step = step_context.step

    emitted_result_names = set()
    for step_output in _yield_compute_results(step_context, inputs, compute_fn, compute_context):
        yield step_output
        if isinstance(step_output, (DynamicOutput, Output)):
            emitted_result_names.add(step_output.output_name)
        elif isinstance(step_output, AssetResult):
            asset_key = (
                step_output.asset_key
                or step_context.job_def.asset_layer.get_asset_key_for_node(step_context.node_handle)
            )
            emitted_result_names.add(
                step_context.job_def.asset_layer.asset_graph.get(
                    asset_key
                ).assets_def.get_output_name_for_asset_key(asset_key)
            )
            # Check results embedded in MaterializeResult are counted
            for check_result in step_output.check_results or []:
                key = check_result.to_asset_check_evaluation(step_context).asset_check_key
                output_name = step_context.job_def.asset_layer.get_op_output_name(key)
                emitted_result_names.add(output_name)
        elif isinstance(step_output, AssetCheckResult):
            if step_output.asset_key and step_output.check_name:
                key = AssetCheckKey(step_output.asset_key, step_output.check_name)
            else:
                key = step_output.to_asset_check_evaluation(step_context).asset_check_key
            output_name = step_context.job_def.asset_layer.get_op_output_name(key)
            if output_name:
                emitted_result_names.add(output_name)

    expected_op_output_names = {
        output.name
        for output in step.step_outputs
        # checks are required if we're in requires_typed_event_stream mode
        if step_context.requires_typed_event_stream or output.properties.asset_check_key
    }
    omitted_outputs = expected_op_output_names.intersection(
        step_context.selected_output_names
    ).difference(emitted_result_names)
    if omitted_outputs:
        message = (
            f"{step_context.op_def.node_type_str} '{step.node_handle}' did not yield or return "
            f"expected outputs {omitted_outputs!r}."
        )

        if step_context.requires_typed_event_stream:
            if step_context.typed_event_stream_error_message:
                message += " " + step_context.typed_event_stream_error_message
            raise DagsterInvariantViolationError(message)
        else:
            step_context.log.info(message)



from datetime import datetime
from typing import TYPE_CHECKING, NamedTuple, Optional

import dagster._check as check
from dagster._annotations import public
from dagster._core.definitions.partitions.context import partition_loading_context
from dagster._core.definitions.partitions.definition.partitions_definition import (
    PartitionsDefinition,
)
from dagster._core.definitions.partitions.mapping.partition_mapping import (
    PartitionMapping,
    UpstreamPartitionsResult,
)
from dagster._core.definitions.partitions.subset.all import AllPartitionsSubset
from dagster._core.definitions.partitions.subset.partitions_subset import PartitionsSubset
from dagster._serdes import whitelist_for_serdes

if TYPE_CHECKING:
    from dagster._core.instance import DynamicPartitionsStore


@whitelist_for_serdes
@public
class AllPartitionMapping(PartitionMapping, NamedTuple("_AllPartitionMapping", [])):
    """Maps every partition in the downstream asset to every partition in the upstream asset.

    Commonly used in the case when the downstream asset is not partitioned, in which the entire
    downstream asset depends on all partitions of the upstream asset.
    """

    def validate_partition_mapping(
        self,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition | None,
    ):
        pass

    def get_upstream_mapped_partitions_result_for_partitions(
        self,
        downstream_partitions_subset: PartitionsSubset | None,
        downstream_partitions_def: PartitionsDefinition | None,
        upstream_partitions_def: PartitionsDefinition,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> UpstreamPartitionsResult:
        with partition_loading_context(current_time, dynamic_partitions_store) as ctx:
            if ctx.dynamic_partitions_store is not None and ctx.effective_dt is not None:
                partitions_subset = AllPartitionsSubset(
                    partitions_def=upstream_partitions_def, context=ctx
                )
            else:
                partitions_subset = upstream_partitions_def.subset_with_all_partitions()
            return UpstreamPartitionsResult(
                partitions_subset=partitions_subset,
                required_but_nonexistent_subset=upstream_partitions_def.empty_subset(),
            )

    def get_downstream_partitions_for_partitions(
        self,
        upstream_partitions_subset: PartitionsSubset,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> PartitionsSubset:
        with partition_loading_context(current_time, dynamic_partitions_store):
            if upstream_partitions_subset is None:
                check.failed("upstream asset is not partitioned")

            if len(upstream_partitions_subset) == 0:
                return downstream_partitions_def.empty_subset()

            return downstream_partitions_def.subset_with_all_partitions()

    @property
    def description(self) -> str:
        return "Each downstream partition depends on all partitions of the upstream asset."



import warnings
from functools import lru_cache
from typing import TYPE_CHECKING, Optional

from dagster._utils.warnings import disable_dagster_warnings

if TYPE_CHECKING:
    from dagster._core.definitions.partitions.definition import PartitionsDefinition
    from dagster._core.definitions.partitions.mapping import PartitionMapping


def infer_partition_mapping(
    partition_mapping: Optional["PartitionMapping"],
    downstream_partitions_def: Optional["PartitionsDefinition"],
    upstream_partitions_def: Optional["PartitionsDefinition"],
) -> "PartitionMapping":
    from dagster._core.definitions.partitions.definition import TimeWindowPartitionsDefinition
    from dagster._core.definitions.partitions.mapping import (
        AllPartitionMapping,
        IdentityPartitionMapping,
        TimeWindowPartitionMapping,
    )
    from dagster._core.definitions.partitions.mapping.multi.multi_to_single import (
        MultiToSingleDimensionPartitionMapping,
        get_infer_single_to_multi_dimension_deps_result,
    )

    if partition_mapping is not None:
        return partition_mapping
    elif upstream_partitions_def and downstream_partitions_def:
        if get_infer_single_to_multi_dimension_deps_result(
            upstream_partitions_def, downstream_partitions_def
        ).can_infer:
            with disable_dagster_warnings():
                return MultiToSingleDimensionPartitionMapping()
        elif isinstance(upstream_partitions_def, TimeWindowPartitionsDefinition) and isinstance(
            downstream_partitions_def, TimeWindowPartitionsDefinition
        ):
            return TimeWindowPartitionMapping()
        else:
            return IdentityPartitionMapping()
    else:
        return AllPartitionMapping()


@lru_cache(maxsize=1)
def get_builtin_partition_mapping_types() -> tuple[type["PartitionMapping"], ...]:
    from dagster._core.definitions.partitions.mapping import (
        AllPartitionMapping,
        IdentityPartitionMapping,
        LastPartitionMapping,
        MultiPartitionMapping,
        MultiToSingleDimensionPartitionMapping,
        SpecificPartitionsPartitionMapping,
        StaticPartitionMapping,
        TimeWindowPartitionMapping,
    )

    return (
        AllPartitionMapping,
        IdentityPartitionMapping,
        LastPartitionMapping,
        SpecificPartitionsPartitionMapping,
        StaticPartitionMapping,
        TimeWindowPartitionMapping,
        MultiToSingleDimensionPartitionMapping,
        MultiPartitionMapping,
    )


def warn_if_partition_mapping_not_builtin(partition_mapping: "PartitionMapping") -> None:
    builtin_partition_mappings = get_builtin_partition_mapping_types()
    if not isinstance(partition_mapping, builtin_partition_mappings):
        warnings.warn(
            f"Non-built-in PartitionMappings, such as {type(partition_mapping).__name__} "
            "are deprecated and will not work with asset reconciliation. The built-in "
            "partition mappings are "
            + ", ".join(
                builtin_partition_mapping.__name__
                for builtin_partition_mapping in builtin_partition_mappings
            )
            + ".",
            category=DeprecationWarning,
        )



from collections.abc import Sequence
from datetime import datetime

import dagster._check as check
from dagster._annotations import public
from dagster._core.definitions.partitions.definition.time_window import (
    TimeWindowPartitionsDefinition,
)
from dagster._core.definitions.partitions.schedule_type import ScheduleType
from dagster._core.definitions.timestamp import TimestampWithTimezone
from dagster._utils.partitions import DEFAULT_DATE_FORMAT, DEFAULT_HOURLY_FORMAT_WITHOUT_TIMEZONE


@public
class HourlyPartitionsDefinition(TimeWindowPartitionsDefinition):
    """A set of hourly partitions.

    The first partition in the set will start on the start_date at midnight. The last partition
    in the set will end before the current time, unless the end_offset argument is set to a
    positive number. If minute_offset is provided, the start and end times of each partition
    will be minute_offset past the hour.

    Args:
        start_date (Union[datetime.datetime, str]): The first date in the set of partitions. Can
            provide in either a datetime or string format.
        end_date (Union[datetime.datetime, str, None]): The last date(excluding) in the set of partitions.
            Default is None. Can provide in either a datetime or string format.
        minute_offset (int): Number of minutes past the hour to "split" the partition. Defaults
            to 0.
        fmt (Optional[str]): The date format to use. Defaults to `%Y-%m-%d`. Note that if a non-UTC
            timezone is used, the date format must include a timezone offset to disambiguate between
            multiple instances of the same time before and after the Fall DST transition. If the
            format does not contain this offset, the second instance of the ambiguous time partition
            key will have the UTC offset automatically appended to it.
        timezone (Optional[str]): The timezone in which each date should exist.
            Supported strings for timezones are the ones provided by the
            `IANA time zone database <https://www.iana.org/time-zones>`_ - e.g. "America/Los_Angeles".
        end_offset (int): Extends the partition set by a number of partitions equal to the value
            passed. If end_offset is 0 (the default), the last partition ends before the current
            time. If end_offset is 1, the second-to-last partition ends before the current time,
            and so on.
        exclusions (Optional[Sequence[Union[str, datetime]]]): Specifies a sequence of cron strings
            or datetime objects that should be excluded from the partition set. Every tick of the
            cron schedule that matches an excluded datetime or matches the tick of an excluded
            cron string will be excluded from the partition set.

    .. code-block:: python

        from datetime import datetime
        from dagster import HourlyPartitionsDefinition

        # Basic hourly partitions starting at midnight
        hourly_partitions = HourlyPartitionsDefinition(start_date=datetime(2022, 3, 12))

        # Hourly partitions with 15-minute offset
        offset_partitions = HourlyPartitionsDefinition(
            start_date=datetime(2022, 3, 12),
            minute_offset=15
        )
    """

    # mapping for fields defined on TimeWindowPartitionsDefinition to this subclasses __new__
    __field_remap__ = {
        "start_ts": "start_date",
        "end_ts": "end_date",
    }

    def __new__(
        cls,
        start_date: datetime | str,
        end_date: datetime | str | None = None,
        minute_offset: int = 0,
        timezone: str | None = None,
        fmt: str | None = None,
        end_offset: int = 0,
        exclusions: Sequence[str | datetime | TimestampWithTimezone] | None = None,
        **kwargs,
    ):
        _fmt = fmt or DEFAULT_HOURLY_FORMAT_WITHOUT_TIMEZONE
        schedule_type = ScheduleType.HOURLY

        # We accept cron_schedule "hidden" via kwargs to support record copy()
        cron_schedule = kwargs.get("cron_schedule")
        if cron_schedule:
            schedule_type = None

        return super().__new__(
            cls,
            schedule_type=schedule_type,
            start=start_date,
            end=end_date,
            minute_offset=minute_offset,
            timezone=timezone,
            fmt=_fmt,
            end_offset=end_offset,
            cron_schedule=cron_schedule,
            exclusions=exclusions,
        )


@public
class DailyPartitionsDefinition(TimeWindowPartitionsDefinition):
    """A set of daily partitions.

    The first partition in the set will start at the start_date at midnight. The last partition
    in the set will end before the current time, unless the end_offset argument is set to a
    positive number. If minute_offset and/or hour_offset are used, the start and end times of
    each partition will be hour_offset:minute_offset of each day.

    Args:
        start_date (Union[datetime.datetime, str]): The first date in the set of partitions. Can
            provide in either a datetime or string format.
        end_date (Union[datetime.datetime, str, None]): The last date(excluding) in the set of partitions.
            Default is None. Can provide in either a datetime or string format.
        minute_offset (int): Number of minutes past the hour to "split" the partition. Defaults
            to 0.
        hour_offset (int): Number of hours past 00:00 to "split" the partition. Defaults to 0.
        timezone (Optional[str]): The timezone in which each date should exist.
            Supported strings for timezones are the ones provided by the
            `IANA time zone database <https://www.iana.org/time-zones>`_ - e.g. "America/Los_Angeles".
        fmt (Optional[str]): The date format to use. Defaults to `%Y-%m-%d`.
        end_offset (int): Extends the partition set by a number of partitions equal to the value
            passed. If end_offset is 0 (the default), the last partition ends before the current
            time. If end_offset is 1, the second-to-last partition ends before the current time,
            and so on.
        exclusions (Optional[Sequence[Union[str, datetime]]]): Specifies a sequence of cron strings
            or datetime objects that should be excluded from the partition set. Every tick of the
            cron schedule that matches an excluded datetime or matches the tick of an excluded
            cron string will be excluded from the partition set.

    .. code-block:: python

        DailyPartitionsDefinition(start_date="2022-03-12")
        # creates partitions (2022-03-12-00:00, 2022-03-13-00:00), (2022-03-13-00:00, 2022-03-14-00:00), ...

        DailyPartitionsDefinition(start_date="2022-03-12", minute_offset=15, hour_offset=16)
        # creates partitions (2022-03-12-16:15, 2022-03-13-16:15), (2022-03-13-16:15, 2022-03-14-16:15), ...
    """

    # mapping for fields defined on TimeWindowPartitionsDefinition to this subclasses __new__
    __field_remap__ = {
        "start_ts": "start_date",
        "end_ts": "end_date",
    }

    def __new__(
        cls,
        start_date: datetime | str,
        end_date: datetime | str | None = None,
        minute_offset: int = 0,
        hour_offset: int = 0,
        timezone: str | None = None,
        fmt: str | None = None,
        end_offset: int = 0,
        exclusions: Sequence[str | datetime | TimestampWithTimezone] | None = None,
        **kwargs,
    ):
        _fmt = fmt or DEFAULT_DATE_FORMAT

        schedule_type = ScheduleType.DAILY

        # We accept cron_schedule "hidden" via kwargs to support record copy()
        cron_schedule = kwargs.get("cron_schedule")
        if cron_schedule:
            schedule_type = None

        return super().__new__(
            cls,
            schedule_type=schedule_type,
            start=start_date,
            end=end_date,
            minute_offset=minute_offset,
            hour_offset=hour_offset,
            timezone=timezone,
            fmt=_fmt,
            end_offset=end_offset,
            cron_schedule=cron_schedule,
            exclusions=exclusions,
        )


class WeeklyPartitionsDefinition(TimeWindowPartitionsDefinition):
    """Defines a set of weekly partitions.

    The first partition in the set will start at the start_date. The last partition in the set will
    end before the current time, unless the end_offset argument is set to a positive number. If
    day_offset is provided, the start and end date of each partition will be day of the week
    corresponding to day_offset (0 indexed with Sunday as the start of the week). If
    minute_offset and/or hour_offset are used, the start and end times of each partition will be
    hour_offset:minute_offset of each day.

    Args:
        start_date (Union[datetime.datetime, str]): The first date in the set of partitions will
            Sunday at midnight following start_date. Can provide in either a datetime or string
            format.
        end_date (Union[datetime.datetime, str, None]): The last date(excluding) in the set of partitions.
            Default is None. Can provide in either a datetime or string format.
        minute_offset (int): Number of minutes past the hour to "split" the partition. Defaults
            to 0.
        hour_offset (int): Number of hours past 00:00 to "split" the partition. Defaults to 0.
        day_offset (int): Day of the week to "split" the partition. Defaults to 0 (Sunday).
        timezone (Optional[str]): The timezone in which each date should exist.
            Supported strings for timezones are the ones provided by the
            `IANA time zone database <https://www.iana.org/time-zones>`_ - e.g. "America/Los_Angeles".
        fmt (Optional[str]): The date format to use. Defaults to `%Y-%m-%d`.
        end_offset (int): Extends the partition set by a number of partitions equal to the value
            passed. If end_offset is 0 (the default), the last partition ends before the current
            time. If end_offset is 1, the second-to-last partition ends before the current time,
            and so on.
        exclusions (Optional[Sequence[Union[str, datetime]]]): Specifies a sequence of cron strings
            or datetime objects that should be excluded from the partition set. Every tick of the
            cron schedule that matches an excluded datetime or matches the tick of an excluded
            cron string will be excluded from the partition set.

    .. code-block:: python

        WeeklyPartitionsDefinition(start_date="2022-03-12")
        # creates partitions (2022-03-13-00:00, 2022-03-20-00:00), (2022-03-20-00:00, 2022-03-27-00:00), ...

        WeeklyPartitionsDefinition(start_date="2022-03-12", minute_offset=15, hour_offset=3, day_offset=6)
        # creates partitions (2022-03-12-03:15, 2022-03-19-03:15), (2022-03-19-03:15, 2022-03-26-03:15), ...
    """

    # mapping for fields defined on TimeWindowPartitionsDefinition to this subclasses __new__
    __field_remap__ = {
        "start_ts": "start_date",
        "end_ts": "end_date",
    }

    def __new__(
        cls,
        start_date: datetime | str,
        end_date: datetime | str | None = None,
        minute_offset: int = 0,
        hour_offset: int = 0,
        day_offset: int = 0,
        timezone: str | None = None,
        fmt: str | None = None,
        end_offset: int = 0,
        exclusions: Sequence[str | datetime | TimestampWithTimezone] | None = None,
        **kwargs,
    ):
        _fmt = fmt or DEFAULT_DATE_FORMAT
        schedule_type = ScheduleType.WEEKLY
        # We accept cron_schedule "hidden" via kwargs to support record copy()
        cron_schedule = kwargs.get("cron_schedule")
        if cron_schedule:
            schedule_type = None

        return super().__new__(
            cls,
            schedule_type=schedule_type,
            start=start_date,
            end=end_date,
            minute_offset=minute_offset,
            hour_offset=hour_offset,
            day_offset=day_offset,
            timezone=timezone,
            fmt=_fmt,
            end_offset=end_offset,
            cron_schedule=cron_schedule,
            exclusions=exclusions,
        )


@public
class MonthlyPartitionsDefinition(TimeWindowPartitionsDefinition):
    """A set of monthly partitions.

    The first partition in the set will start at the soonest first of the month after start_date
    at midnight. The last partition in the set will end before the current time, unless the
    end_offset argument is set to a positive number. If day_offset is provided, the start and
    end date of each partition will be day_offset. If minute_offset and/or hour_offset are used,
    the start and end times of each partition will be hour_offset:minute_offset of each day.

    Args:
        start_date (Union[datetime.datetime, str]): The first date in the set of partitions will be
            midnight the soonest first of the month following start_date. Can provide in either a
            datetime or string format.
        end_date (Union[datetime.datetime, str, None]): The last date(excluding) in the set of partitions.
            Default is None. Can provide in either a datetime or string format.
        minute_offset (int): Number of minutes past the hour to "split" the partition. Defaults
            to 0.
        hour_offset (int): Number of hours past 00:00 to "split" the partition. Defaults to 0.
        day_offset (int): Day of the month to "split" the partition. Defaults to 1.
        timezone (Optional[str]): The timezone in which each date should exist.
            Supported strings for timezones are the ones provided by the
            `IANA time zone database <https://www.iana.org/time-zones>`_ - e.g. "America/Los_Angeles".
        fmt (Optional[str]): The date format to use. Defaults to `%Y-%m-%d`.
        end_offset (int): Extends the partition set by a number of partitions equal to the value
            passed. If end_offset is 0 (the default), the last partition ends before the current
            time. If end_offset is 1, the second-to-last partition ends before the current time,
            and so on.
        exclusions (Optional[Sequence[Union[str, datetime]]]): Specifies a sequence of cron strings
            or datetime objects that should be excluded from the partition set. Every tick of the
            cron schedule that matches an excluded datetime or matches the tick of an excluded
            cron string will be excluded from the partition set.

    .. code-block:: python

        MonthlyPartitionsDefinition(start_date="2022-03-12")
        # creates partitions (2022-04-01-00:00, 2022-05-01-00:00), (2022-05-01-00:00, 2022-06-01-00:00), ...

        MonthlyPartitionsDefinition(start_date="2022-03-12", minute_offset=15, hour_offset=3, day_offset=5)
        # creates partitions (2022-04-05-03:15, 2022-05-05-03:15), (2022-05-05-03:15, 2022-06-05-03:15), ...
    """

    # mapping for fields defined on TimeWindowPartitionsDefinition to this subclasses __new__
    __field_remap__ = {
        "start_ts": "start_date",
        "end_ts": "end_date",
    }

    def __new__(
        cls,
        start_date: datetime | str,
        end_date: datetime | str | None = None,
        minute_offset: int = 0,
        hour_offset: int = 0,
        day_offset: int = 1,
        timezone: str | None = None,
        fmt: str | None = None,
        end_offset: int = 0,
        exclusions: Sequence[str | datetime | TimestampWithTimezone] | None = None,
        **kwargs,
    ):
        _fmt = fmt or DEFAULT_DATE_FORMAT
        schedule_type = ScheduleType.MONTHLY

        # We accept cron_schedule "hidden" via kwargs to support record copy()
        cron_schedule = kwargs.get("cron_schedule")
        if cron_schedule:
            schedule_type = None
            check.invariant(
                day_offset == 1,
                f"Reconstruction by cron_schedule got unexpected day_offset {day_offset}, expected 1.",
            )
            day_offset = 0

        return super().__new__(
            cls,
            schedule_type=schedule_type,
            start=start_date,
            end=end_date,
            minute_offset=minute_offset,
            hour_offset=hour_offset,
            day_offset=day_offset,
            timezone=timezone,
            fmt=_fmt,
            end_offset=end_offset,
            cron_schedule=cron_schedule,
            exclusions=exclusions,
        )



import enum

from dagster_shared.serdes import whitelist_for_serdes


@whitelist_for_serdes
class AssetHealthStatus(enum.Enum):
    """Enum for the health status of an asset."""

    HEALTHY = "HEALTHY"
    WARNING = "WARNING"
    DEGRADED = "DEGRADED"
    UNKNOWN = "UNKNOWN"
    NOT_APPLICABLE = "NOT_APPLICABLE"


def overall_status_from_component_statuses(
    asset_checks_status: AssetHealthStatus,
    materialization_status: AssetHealthStatus,
    freshness_status: AssetHealthStatus,
) -> AssetHealthStatus:
    statuses = [
        materialization_status,
        asset_checks_status,
        freshness_status,
    ]
    if AssetHealthStatus.DEGRADED in statuses:
        return AssetHealthStatus.DEGRADED
    if AssetHealthStatus.WARNING in statuses:
        return AssetHealthStatus.WARNING

    # at this point, all statuses are HEALTHY, UNKNOWN, or NOT_APPLICABLE
    # if the materialization status is UNKNOWN, then we want overall status to be UNKNOWN,
    # even if other statuses are known
    if materialization_status == AssetHealthStatus.UNKNOWN:
        return AssetHealthStatus.UNKNOWN
    if all(
        status == AssetHealthStatus.UNKNOWN or status == AssetHealthStatus.NOT_APPLICABLE
        for status in statuses
    ):
        return AssetHealthStatus.UNKNOWN
    # at least one status must be HEALTHY
    return AssetHealthStatus.HEALTHY



import os
import weakref
from collections import defaultdict
from collections.abc import Callable, Iterable, Mapping, Sequence
from contextlib import ExitStack
from types import TracebackType
from typing import TYPE_CHECKING, Any, Optional, Union, cast

if TYPE_CHECKING:
    from tempfile import TemporaryDirectory

from typing import TypeAlias

import yaml
from typing_extensions import Self

import dagster._check as check
from dagster._annotations import public
from dagster._core.instance.config import DAGSTER_CONFIG_YAML_FILENAME
from dagster._core.instance.methods.asset_methods import AssetMethods
from dagster._core.instance.methods.daemon_methods import DaemonMethods
from dagster._core.instance.methods.event_methods import EventMethods
from dagster._core.instance.methods.run_launcher_methods import RunLauncherMethods
from dagster._core.instance.methods.run_methods import RunMethods
from dagster._core.instance.methods.scheduling_methods import SchedulingMethods
from dagster._core.instance.methods.settings_methods import SettingsMethods
from dagster._core.instance.methods.storage_methods import StorageMethods
from dagster._core.instance.ref import InstanceRef
from dagster._core.instance.types import DynamicPartitionsStore, InstanceType
from dagster._core.storage.dagster_run import DagsterRun, RunsFilter
from dagster._serdes import ConfigurableClass
from dagster._utils import PrintFn, traced

if TYPE_CHECKING:
    from dagster._core.debug import DebugRunPayload
    from dagster._core.definitions.asset_checks.asset_check_evaluation import AssetCheckEvaluation
    from dagster._core.definitions.events import AssetKey, AssetObservation
    from dagster._core.definitions.freshness import FreshnessStateChange, FreshnessStateEvaluation
    from dagster._core.definitions.partitions.definition import PartitionsDefinition
    from dagster._core.event_api import AssetRecordsFilter, RunStatusChangeRecordsFilter
    from dagster._core.events import AssetMaterialization, DagsterEventType
    from dagster._core.events.log import EventLogEntry
    from dagster._core.launcher import RunLauncher
    from dagster._core.remote_representation.historical import HistoricalJob
    from dagster._core.run_coordinator import RunCoordinator
    from dagster._core.scheduler import Scheduler
    from dagster._core.secrets import SecretsLoader
    from dagster._core.snap import ExecutionPlanSnapshot, JobSnap
    from dagster._core.storage.compute_log_manager import ComputeLogManager
    from dagster._core.storage.daemon_cursor import DaemonCursorStorage
    from dagster._core.storage.dagster_run import JobBucket, RunRecord, TagBucket
    from dagster._core.storage.defs_state.base import DefsStateStorage
    from dagster._core.storage.event_log import EventLogStorage
    from dagster._core.storage.event_log.base import (
        AssetRecord,
        EventLogConnection,
        EventRecordsResult,
    )
    from dagster._core.storage.partition_status_cache import AssetPartitionStatus
    from dagster._core.storage.root import LocalArtifactStorage
    from dagster._core.storage.runs import RunStorage
    from dagster._core.storage.schedules import ScheduleStorage
    from dagster._core.storage.sql import AlembicVersion

DagsterInstanceOverrides: TypeAlias = Mapping[str, Any]


@public
class DagsterInstance(
    SettingsMethods,
    StorageMethods,
    DaemonMethods,
    RunLauncherMethods,
    EventMethods,
    SchedulingMethods,
    AssetMethods,
    RunMethods,
    DynamicPartitionsStore,
):
    """Core abstraction for managing Dagster's access to storage and other resources.

    Use DagsterInstance.get() to grab the current DagsterInstance which will load based on
    the values in the ``dagster.yaml`` file in ``$DAGSTER_HOME``.

    Alternatively, DagsterInstance.ephemeral() can use used which provides a set of
    transient in-memory components.

    Configuration of this class should be done by setting values in ``$DAGSTER_HOME/dagster.yaml``.
    For example, to use Postgres for dagster storage, you can write a ``dagster.yaml`` such as the
    following:

    .. literalinclude:: ../../../../../examples/docs_snippets/docs_snippets/deploying/dagster-pg.yaml
       :caption: dagster.yaml
       :language: YAML

    Args:
        instance_type (InstanceType): Indicates whether the instance is ephemeral or persistent.
            Users should not attempt to set this value directly or in their ``dagster.yaml`` files.
        local_artifact_storage (LocalArtifactStorage): The local artifact storage is used to
            configure storage for any artifacts that require a local disk, such as schedules, or
            when using the filesystem system storage to manage files and intermediates. By default,
            this will be a :py:class:`dagster._core.storage.root.LocalArtifactStorage`. Configurable
            in ``dagster.yaml`` using the :py:class:`~dagster.serdes.ConfigurableClass`
            machinery.
        run_storage (RunStorage): The run storage is used to store metadata about ongoing and past
            pipeline runs. By default, this will be a
            :py:class:`dagster._core.storage.runs.SqliteRunStorage`. Configurable in ``dagster.yaml``
            using the :py:class:`~dagster.serdes.ConfigurableClass` machinery.
        event_storage (EventLogStorage): Used to store the structured event logs generated by
            pipeline runs. By default, this will be a
            :py:class:`dagster._core.storage.event_log.SqliteEventLogStorage`. Configurable in
            ``dagster.yaml`` using the :py:class:`~dagster.serdes.ConfigurableClass` machinery.
        compute_log_manager (Optional[ComputeLogManager]): The compute log manager handles stdout
            and stderr logging for op compute functions. By default, this will be a
            :py:class:`dagster._core.storage.local_compute_log_manager.LocalComputeLogManager`.
            Configurable in ``dagster.yaml`` using the
            :py:class:`~dagster.serdes.ConfigurableClass` machinery.
        run_coordinator (Optional[RunCoordinator]): A runs coordinator may be used to manage the execution
            of pipeline runs.
        run_launcher (Optional[RunLauncher]): Optionally, a run launcher may be used to enable
            a Dagster instance to launch pipeline runs, e.g. on a remote Kubernetes cluster, in
            addition to running them locally.
        settings (Optional[Dict]): Specifies certain per-instance settings,
            such as feature flags. These are set in the ``dagster.yaml`` under a set of whitelisted
            keys.
        ref (Optional[InstanceRef]): Used by internal machinery to pass instances across process
            boundaries.
    """

    # Stores TemporaryDirectory instances that were created for DagsterInstance.local_temp() calls
    # to be removed once the instance is garbage collected.
    _TEMP_DIRS: "weakref.WeakKeyDictionary[DagsterInstance, TemporaryDirectory]" = (
        weakref.WeakKeyDictionary()
    )

    def __init__(
        self,
        instance_type: InstanceType,
        local_artifact_storage: "LocalArtifactStorage",
        run_storage: "RunStorage",
        event_storage: "EventLogStorage",
        run_coordinator: Optional["RunCoordinator"],
        compute_log_manager: Optional["ComputeLogManager"],
        run_launcher: Optional["RunLauncher"],
        scheduler: Optional["Scheduler"] = None,
        schedule_storage: Optional["ScheduleStorage"] = None,
        settings: Mapping[str, Any] | None = None,
        secrets_loader: Optional["SecretsLoader"] = None,
        defs_state_storage: Optional["DefsStateStorage"] = None,
        ref: InstanceRef | None = None,
        **_kwargs: Any,  # we accept kwargs for forward-compat of custom instances
    ):
        from dagster._core.launcher import RunLauncher
        from dagster._core.run_coordinator import RunCoordinator
        from dagster._core.scheduler import Scheduler
        from dagster._core.secrets import SecretsLoader
        from dagster._core.storage.compute_log_manager import ComputeLogManager
        from dagster._core.storage.defs_state import DefsStateStorage
        from dagster._core.storage.event_log import EventLogStorage
        from dagster._core.storage.root import LocalArtifactStorage
        from dagster._core.storage.runs import RunStorage
        from dagster._core.storage.schedules import ScheduleStorage

        self._instance_type = check.inst_param(instance_type, "instance_type", InstanceType)
        self._local_artifact_storage = check.inst_param(
            local_artifact_storage, "local_artifact_storage", LocalArtifactStorage
        )
        self._event_storage = check.inst_param(event_storage, "event_storage", EventLogStorage)
        self._event_storage.register_instance(self)

        self._run_storage = check.inst_param(run_storage, "run_storage", RunStorage)
        self._run_storage.register_instance(self)

        if compute_log_manager:
            self._compute_log_manager = check.inst_param(
                compute_log_manager, "compute_log_manager", ComputeLogManager
            )
            self._compute_log_manager.register_instance(self)
        else:
            check.invariant(
                ref,
                "Compute log manager must be provided if instance is not from a ref",
            )
            self._compute_log_manager = None

        self._scheduler = check.opt_inst_param(scheduler, "scheduler", Scheduler)

        self._schedule_storage = check.opt_inst_param(
            schedule_storage, "schedule_storage", ScheduleStorage
        )
        if self._schedule_storage:
            self._schedule_storage.register_instance(self)

        if run_coordinator:
            self._run_coordinator = check.inst_param(
                run_coordinator, "run_coordinator", RunCoordinator
            )
            self._run_coordinator.register_instance(self)
        else:
            check.invariant(ref, "Run coordinator must be provided if instance is not from a ref")
            self._run_coordinator = None

        if run_launcher:
            self._run_launcher: RunLauncher | None = check.inst_param(
                run_launcher, "run_launcher", RunLauncher
            )
            run_launcher.register_instance(self)
        else:
            check.invariant(ref, "Run launcher must be provided if instance is not from a ref")
            self._run_launcher = None

        self._settings = check.opt_mapping_param(settings, "settings")

        self._secrets_loader = check.opt_inst_param(secrets_loader, "secrets_loader", SecretsLoader)

        if self._secrets_loader:
            self._secrets_loader.register_instance(self)

        self._defs_state_storage = check.opt_inst_param(
            defs_state_storage, "defs_state_storage", DefsStateStorage
        )
        if self._defs_state_storage:
            self._defs_state_storage.register_instance(self)

        self._ref = check.opt_inst_param(ref, "ref", InstanceRef)

        self._subscribers: dict[str, list[Callable]] = defaultdict(list)

        self._initialize_run_monitoring()

        # Used for batched event handling
        self._event_buffer: dict[str, list[EventLogEntry]] = defaultdict(list)

    # =====================================================================================
    # PUBLIC API METHODS
    # =====================================================================================

    # Factory Methods
    # ---------------

    @public
    @staticmethod
    def ephemeral(
        tempdir: str | None = None,
        preload: Sequence["DebugRunPayload"] | None = None,
        settings: dict | None = None,
    ) -> "DagsterInstance":
        """Create a `DagsterInstance` suitable for ephemeral execution, useful in test contexts. An
        ephemeral instance uses mostly in-memory components. Use `local_temp` to create a test
        instance that is fully persistent.

        Args:
            tempdir (Optional[str]): The path of a directory to be used for local artifact storage.
            preload (Optional[Sequence[DebugRunPayload]]): A sequence of payloads to load into the
                instance's run storage. Useful for debugging.
            settings (Optional[Dict]): Settings for the instance.

        Returns:
            DagsterInstance: An ephemeral DagsterInstance.
        """
        from dagster._core.instance.factory import create_ephemeral_instance

        return create_ephemeral_instance(tempdir=tempdir, preload=preload, settings=settings)

    @public
    @staticmethod
    def get() -> "DagsterInstance":
        """Get the current `DagsterInstance` as specified by the ``DAGSTER_HOME`` environment variable.

        Returns:
            DagsterInstance: The current DagsterInstance.
        """
        from dagster._core.instance.factory import create_instance_from_dagster_home

        return create_instance_from_dagster_home()

    @public
    @staticmethod
    def local_temp(
        tempdir: str | None = None,
        overrides: DagsterInstanceOverrides | None = None,
    ) -> "DagsterInstance":
        """Create a DagsterInstance that uses a temporary directory for local storage. This is a
        regular, fully persistent instance. Use `ephemeral` to get an ephemeral instance with
        in-memory components.

        Args:
            tempdir (Optional[str]): The path of a directory to be used for local artifact storage.
            overrides (Optional[DagsterInstanceOverrides]): Override settings for the instance.

        Returns:
            DagsterInstance
        """
        from dagster._core.instance.factory import create_local_temp_instance

        return create_local_temp_instance(tempdir=tempdir, overrides=overrides)

    # Asset Domain
    # ------------

    @public
    @traced
    def fetch_materializations(
        self,
        records_filter: Union["AssetKey", "AssetRecordsFilter"],
        limit: int,
        cursor: str | None = None,
        ascending: bool = False,
    ) -> "EventRecordsResult":
        """Return a list of materialization records stored in the event log storage.

        Args:
            records_filter (Union[AssetKey, AssetRecordsFilter]): the filter by which to
                filter event records.
            limit (int): Number of results to get.
            cursor (Optional[str]): Cursor to use for pagination. Defaults to None.
            ascending (Optional[bool]): Sort the result in ascending order if True, descending
                otherwise. Defaults to descending.

        Returns:
            EventRecordsResult: Object containing a list of event log records and a cursor string
        """
        return AssetMethods.fetch_materializations(self, records_filter, limit, cursor, ascending)

    @public
    @traced
    def fetch_observations(
        self,
        records_filter: Union["AssetKey", "AssetRecordsFilter"],
        limit: int,
        cursor: str | None = None,
        ascending: bool = False,
    ) -> "EventRecordsResult":
        """Return a list of observation records stored in the event log storage.

        Args:
            records_filter (Optional[Union[AssetKey, AssetRecordsFilter]]): the filter by which to
                filter event records.
            limit (int): Number of results to get.
            cursor (Optional[str]): Cursor to use for pagination. Defaults to None.
            ascending (Optional[bool]): Sort the result in ascending order if True, descending
                otherwise. Defaults to descending.

        Returns:
            EventRecordsResult: Object containing a list of event log records and a cursor string
        """
        return self._event_storage.fetch_observations(records_filter, limit, cursor, ascending)

    @public
    @traced
    def get_asset_keys(
        self,
        prefix: Sequence[str] | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> Sequence["AssetKey"]:
        """Return a filtered subset of asset keys managed by this instance.

        Args:
            prefix (Optional[Sequence[str]]): Return only assets having this key prefix.
            limit (Optional[int]): Maximum number of keys to return.
            cursor (Optional[str]): Cursor to use for pagination.

        Returns:
            Sequence[AssetKey]: List of asset keys.
        """
        return AssetMethods.get_asset_keys(self, prefix, limit, cursor)

    @public
    @traced
    def get_asset_records(
        self, asset_keys: Sequence["AssetKey"] | None = None
    ) -> Sequence["AssetRecord"]:
        """Return an `AssetRecord` for each of the given asset keys.

        Args:
            asset_keys (Optional[Sequence[AssetKey]]): List of asset keys to retrieve records for.

        Returns:
            Sequence[AssetRecord]: List of asset records.
        """
        return AssetMethods.get_asset_records(self, asset_keys)

    @public
    def get_latest_materialization_code_versions(
        self, asset_keys: Iterable["AssetKey"]
    ) -> Mapping["AssetKey", str | None]:
        """Returns the code version used for the latest materialization of each of the provided
        assets.

        Args:
            asset_keys (Iterable[AssetKey]): The asset keys to find latest materialization code
                versions for.

        Returns:
            Mapping[AssetKey, Optional[str]]: A dictionary with a key for each of the provided asset
                keys. The values will be None if the asset has no materializations. If an asset does
                not have a code version explicitly assigned to its definitions, but was
                materialized, Dagster assigns the run ID as its code version.
        """
        return AssetMethods.get_latest_materialization_code_versions(self, asset_keys)

    @public
    @traced
    def get_latest_materialization_event(self, asset_key: "AssetKey") -> Optional["EventLogEntry"]:
        """Fetch the latest materialization event for the given asset key.

        Args:
            asset_key (AssetKey): Asset key to return materialization for.

        Returns:
            Optional[EventLogEntry]: The latest materialization event for the given asset
                key, or `None` if the asset has not been materialized.
        """
        return AssetMethods.get_latest_materialization_event(self, asset_key)

    @public
    @traced
    def get_status_by_partition(
        self,
        asset_key: "AssetKey",
        partition_keys: Sequence[str],
        partitions_def: "PartitionsDefinition",
    ) -> Mapping[str, "AssetPartitionStatus"] | None:
        """Get the current status of provided partition_keys for the provided asset.

        Args:
            asset_key (AssetKey): The asset to get per-partition status for.
            partition_keys (Sequence[str]): The partitions to get status for.
            partitions_def (PartitionsDefinition): The PartitionsDefinition of the asset to get
                per-partition status for.

        Returns:
            Optional[Mapping[str, AssetPartitionStatus]]: status for each partition key

        """
        return AssetMethods.get_status_by_partition(self, asset_key, partition_keys, partitions_def)

    @public
    @traced
    def has_asset_key(self, asset_key: "AssetKey") -> bool:
        """Return true if this instance manages the given asset key.

        Args:
            asset_key (AssetKey): Asset key to check.
        """
        return AssetMethods.has_asset_key(self, asset_key)

    @public
    def report_runless_asset_event(
        self,
        asset_event: Union[
            "AssetMaterialization",
            "AssetObservation",
            "AssetCheckEvaluation",
            "FreshnessStateEvaluation",
            "FreshnessStateChange",
        ],
    ):
        """Record an event log entry related to assets that does not belong to a Dagster run."""
        return AssetMethods.report_runless_asset_event(self, asset_event)

    @public
    @traced
    def wipe_assets(self, asset_keys: Sequence["AssetKey"]) -> None:
        """Wipes asset event history from the event log for the given asset keys.

        Args:
            asset_keys (Sequence[AssetKey]): Asset keys to wipe.
        """
        AssetMethods.wipe_assets(self, asset_keys)

    # Run Domain
    # ----------

    @public
    @traced
    def delete_run(self, run_id: str) -> None:
        """Delete a run and all events generated by that from storage.

        Args:
            run_id (str): The id of the run to delete.
        """
        self._run_storage.delete_run(run_id)
        self._event_storage.delete_events(run_id)

    @public
    def get_run_by_id(self, run_id: str) -> DagsterRun | None:
        """Get a :py:class:`DagsterRun` matching the provided `run_id`.

        Args:
            run_id (str): The id of the run to retrieve.

        Returns:
            Optional[DagsterRun]: The run corresponding to the given id. If no run matching the id
                is found, return `None`.
        """
        record = self.get_run_record_by_id(run_id)
        if record is None:
            return None
        return record.dagster_run

    @public
    @traced
    def get_run_record_by_id(self, run_id: str) -> Optional["RunRecord"]:
        """Get a :py:class:`RunRecord` matching the provided `run_id`.

        Args:
            run_id (str): The id of the run record to retrieve.

        Returns:
            Optional[RunRecord]: The run record corresponding to the given id. If no run matching
                the id is found, return `None`.
        """
        if not run_id:
            return None
        records = self._run_storage.get_run_records(RunsFilter(run_ids=[run_id]), limit=1)
        if not records:
            return None
        return records[0]

    @public
    @traced
    def get_run_records(
        self,
        filters: Optional["RunsFilter"] = None,
        limit: int | None = None,
        order_by: str | None = None,
        ascending: bool = False,
        cursor: str | None = None,
        bucket_by: Union["JobBucket", "TagBucket"] | None = None,
    ) -> Sequence["RunRecord"]:
        """Return a list of run records stored in the run storage, sorted by the given column in given order.

        Args:
            filters (Optional[RunsFilter]): the filter by which to filter runs.
            limit (Optional[int]): Number of results to get. Defaults to infinite.
            order_by (Optional[str]): Name of the column to sort by. Defaults to id.
            ascending (Optional[bool]): Sort the result in ascending order if True, descending
                otherwise. Defaults to descending.

        Returns:
            List[RunRecord]: List of run records stored in the run storage.
        """
        return self._run_storage.get_run_records(
            filters, limit, order_by, ascending, cursor, bucket_by
        )

    # Event Domain
    # ------------

    @public
    @traced
    def fetch_run_status_changes(
        self,
        records_filter: Union["DagsterEventType", "RunStatusChangeRecordsFilter"],
        limit: int,
        cursor: str | None = None,
        ascending: bool = False,
    ) -> "EventRecordsResult":
        """Return a list of run_status_event records stored in the event log storage.

        Args:
            records_filter (Optional[Union[DagsterEventType, RunStatusChangeRecordsFilter]]): the
                filter by which to filter event records.
            limit (int): Number of results to get.
            cursor (Optional[str]): Cursor to use for pagination. Defaults to None.
            ascending (Optional[bool]): Sort the result in ascending order if True, descending
                otherwise. Defaults to descending.

        Returns:
            EventRecordsResult: Object containing a list of event log records and a cursor string
        """
        return self._event_storage.fetch_run_status_changes(
            records_filter, limit, cursor, ascending
        )

    @traced
    def get_records_for_run(
        self,
        run_id: str,
        cursor: str | None = None,
        of_type: Union["DagsterEventType", set["DagsterEventType"]] | None = None,
        limit: int | None = None,
        ascending: bool = True,
    ) -> "EventLogConnection":
        """Get event records for run.

        NOTE: This method is duplicated here (vs only being in EventMethods) because
        some Datadog tracing spans specifically expect to find this method on the
        DagsterInstance class for proper trace attribution.
        """
        return EventMethods.get_records_for_run(self, run_id, cursor, of_type, limit, ascending)

    # Storage/Partition Domain
    # ------------------------

    @public
    @traced
    def add_dynamic_partitions(
        self, partitions_def_name: str, partition_keys: Sequence[str]
    ) -> None:
        """Add partitions to the specified :py:class:`DynamicPartitionsDefinition` idempotently.
        Does not add any partitions that already exist.

        Args:
            partitions_def_name (str): The name of the `DynamicPartitionsDefinition`.
            partition_keys (Sequence[str]): Partition keys to add.
        """
        return self._event_storage.add_dynamic_partitions(partitions_def_name, partition_keys)

    @public
    @traced
    def delete_dynamic_partition(self, partitions_def_name: str, partition_key: str) -> None:
        """Delete a partition for the specified :py:class:`DynamicPartitionsDefinition`.
        If the partition does not exist, exits silently.

        Args:
            partitions_def_name (str): The name of the `DynamicPartitionsDefinition`.
            partition_key (str): Partition key to delete.
        """
        return self._event_storage.delete_dynamic_partition(partitions_def_name, partition_key)

    @public
    @traced
    def get_dynamic_partitions(self, partitions_def_name: str) -> Sequence[str]:
        """Get the set of partition keys for the specified :py:class:`DynamicPartitionsDefinition`.

        Args:
            partitions_def_name (str): The name of the `DynamicPartitionsDefinition`.
        """
        return self._event_storage.get_dynamic_partitions(partitions_def_name)

    def get_dynamic_partitions_definition_id(self, partitions_def_name: str) -> str:
        from dagster._core.definitions.partitions.context import partition_loading_context
        from dagster._core.definitions.partitions.utils import (
            generate_partition_key_based_definition_id,
        )

        with partition_loading_context() as calling_context:
            dynamic_partitions_store = calling_context.dynamic_partitions_store or self
            # matches the base implementation of the get_serializable_unique_identifier on PartitionsDefinition
            partition_keys = dynamic_partitions_store.get_dynamic_partitions(partitions_def_name)
            return generate_partition_key_based_definition_id(partition_keys)

    @public
    @traced
    def has_dynamic_partition(self, partitions_def_name: str, partition_key: str) -> bool:
        """Check if a partition key exists for the :py:class:`DynamicPartitionsDefinition`.

        Args:
            partitions_def_name (str): The name of the `DynamicPartitionsDefinition`.
            partition_key (str): Partition key to check.
        """
        return self._event_storage.has_dynamic_partition(partitions_def_name, partition_key)

    # =====================================================================================
    # INTERNAL METHODS
    # =====================================================================================

    # Core Instance Methods
    # ---------------------

    @staticmethod
    def from_config(
        config_dir: str,
        config_filename: str = DAGSTER_CONFIG_YAML_FILENAME,
    ) -> "DagsterInstance":
        from dagster._core.instance.factory import create_instance_from_config

        return create_instance_from_config(config_dir, config_filename)

    @staticmethod
    def from_ref(instance_ref: InstanceRef) -> "DagsterInstance":
        from dagster._core.instance.factory import create_instance_from_ref

        return create_instance_from_ref(instance_ref)

    # flags

    @property
    def is_persistent(self) -> bool:
        return self._instance_type == InstanceType.PERSISTENT

    @property
    def is_ephemeral(self) -> bool:
        return self._instance_type == InstanceType.EPHEMERAL

    def get_ref(self) -> InstanceRef:
        if self._ref:
            return self._ref

        check.failed(
            "Attempted to prepare an ineligible DagsterInstance ({inst_type}) for cross "
            "process communication.{dagster_home_msg}".format(
                inst_type=self._instance_type,
                dagster_home_msg=(
                    "\nDAGSTER_HOME environment variable is not set, set it to "
                    "a directory on the filesystem for dagster to use for storage and cross "
                    "process coordination."
                    if os.getenv("DAGSTER_HOME") is None
                    else ""
                ),
            )
        )

    @property
    def root_directory(self) -> str:
        return self._local_artifact_storage.base_dir

    def _info(self, component: object) -> str | Mapping[Any, Any]:
        # ConfigurableClass may not have inst_data if it's a direct instantiation
        # which happens for ephemeral instances
        if isinstance(component, ConfigurableClass) and component.inst_data:
            return component.inst_data.info_dict()
        if type(component) is dict:
            return component
        return component.__class__.__name__

    def _info_str_for_component(self, component_name: str, component: object) -> str:
        return yaml.dump(
            {component_name: self._info(component)},
            default_flow_style=False,
            sort_keys=False,
        )

    def info_dict(self) -> Mapping[str, object]:
        settings: Mapping[str, object] = self._settings if self._settings else {}

        ret = {
            "local_artifact_storage": self._info(self._local_artifact_storage),
            "run_storage": self._info(self._run_storage),
            "event_log_storage": self._info(self._event_storage),
            "compute_logs": self._info(self._compute_log_manager),
            "schedule_storage": self._info(self._schedule_storage),
            "scheduler": self._info(self._scheduler),
            "run_coordinator": self._info(self._run_coordinator),
            "run_launcher": self._info(self.run_launcher),
        }
        ret.update(
            {
                settings_key: self._info(settings_value)
                for settings_key, settings_value in settings.items()
            }
        )

        return ret

    def info_str(self) -> str:
        return yaml.dump(self.info_dict(), default_flow_style=False, sort_keys=False)

    def schema_str(self) -> str:
        def _schema_dict(
            alembic_version: Optional["AlembicVersion"],
        ) -> Mapping[str, object] | None:
            if not alembic_version:
                return None
            db_revision, head_revision = alembic_version
            return {
                "current": db_revision,
                "latest": head_revision,
            }

        return yaml.dump(
            {
                "schema": {
                    "event_log_storage": _schema_dict(self._event_storage.alembic_version()),
                    "run_storage": _schema_dict(self._event_storage.alembic_version()),
                    "schedule_storage": _schema_dict(self._event_storage.alembic_version()),
                }
            },
            default_flow_style=False,
            sort_keys=False,
        )

    @property
    def run_storage(self) -> "RunStorage":
        return self._run_storage

    @property
    def defs_state_storage(self) -> Optional["DefsStateStorage"]:
        return self._defs_state_storage

    @property
    def event_log_storage(self) -> "EventLogStorage":
        return self._event_storage

    @property
    def daemon_cursor_storage(self) -> "DaemonCursorStorage":
        return self._run_storage

    @property
    def scheduler(self) -> Optional["Scheduler"]:
        return self._scheduler

    # run coordinator

    @property
    def run_coordinator(self) -> "RunCoordinator":
        # Lazily load in case the run coordinator requires dependencies that are not available
        # everywhere that loads the instance
        if not self._run_coordinator:
            check.invariant(
                self._ref, "Run coordinator not provided, and no instance ref available"
            )
            run_coordinator = cast("InstanceRef", self._ref).run_coordinator
            check.invariant(run_coordinator, "Run coordinator not configured in instance ref")
            self._run_coordinator = cast("RunCoordinator", run_coordinator)
            self._run_coordinator.register_instance(self)
        return self._run_coordinator

    @property
    def run_launcher(self) -> "RunLauncher":
        # Lazily load in case the launcher requires dependencies that are not available everywhere
        # that loads the instance (e.g. The EcsRunLauncher requires boto3)
        if not self._run_launcher:
            check.invariant(self._ref, "Run launcher not provided, and no instance ref available")
            launcher = cast("InstanceRef", self._ref).run_launcher
            check.invariant(launcher, "Run launcher not configured in instance ref")
            self._run_launcher = cast("RunLauncher", launcher)
            self._run_launcher.register_instance(self)
        return self._run_launcher

    # compute logs

    @property
    def compute_log_manager(self) -> "ComputeLogManager":
        if not self._compute_log_manager:
            check.invariant(
                self._ref,
                "Compute log manager not provided, and no instance ref available",
            )
            compute_log_manager = cast("InstanceRef", self._ref).compute_log_manager
            check.invariant(
                compute_log_manager,
                "Compute log manager not configured in instance ref",
            )
            self._compute_log_manager = cast("ComputeLogManager", compute_log_manager)
            self._compute_log_manager.register_instance(self)
        return self._compute_log_manager

    def get_settings(self, settings_key: str) -> Any:
        check.str_param(settings_key, "settings_key")
        if self._settings and settings_key in self._settings:
            return self._settings.get(settings_key)
        return {}

    def upgrade(self, print_fn: PrintFn | None = None) -> None:
        from dagster._core.storage.migration.utils import upgrading_instance

        with upgrading_instance(self):
            if print_fn:
                print_fn("Updating run storage...")
            self._run_storage.upgrade()  # type: ignore  # (unknown method on run storage)
            self._run_storage.migrate(print_fn)

            if print_fn:
                print_fn("Updating event storage...")
            self._event_storage.upgrade()
            self._event_storage.reindex_assets(print_fn=print_fn)

            if print_fn:
                print_fn("Updating schedule storage...")
            self._schedule_storage.upgrade()  # type: ignore  # (possible none)
            self._schedule_storage.migrate(print_fn)  # type: ignore  # (possible none)

    def dispose(self) -> None:
        StorageMethods.dispose(self)
        if self._run_coordinator:
            self._run_coordinator.dispose()
        if self._run_launcher:
            self._run_launcher.dispose()
        if self._compute_log_manager:
            self._compute_log_manager.dispose()
        if self._secrets_loader:
            self._secrets_loader.dispose()

        if self in DagsterInstance._TEMP_DIRS:
            DagsterInstance._TEMP_DIRS[self].cleanup()
            del DagsterInstance._TEMP_DIRS[self]

    # run storage

    @traced
    def get_job_snapshot(self, snapshot_id: str) -> "JobSnap":
        return self._run_storage.get_job_snapshot(snapshot_id)

    @traced
    def has_job_snapshot(self, snapshot_id: str) -> bool:
        return self._run_storage.has_job_snapshot(snapshot_id)

    @traced
    def has_snapshot(self, snapshot_id: str) -> bool:
        return self._run_storage.has_snapshot(snapshot_id)

    @traced
    def get_historical_job(self, snapshot_id: str) -> "HistoricalJob":
        from dagster._core.remote_representation.historical import HistoricalJob

        snapshot = self._run_storage.get_job_snapshot(snapshot_id)
        parent_snapshot = (
            self._run_storage.get_job_snapshot(snapshot.lineage_snapshot.parent_snapshot_id)
            if snapshot.lineage_snapshot
            else None
        )
        return HistoricalJob(snapshot, snapshot_id, parent_snapshot)

    @traced
    def has_historical_job(self, snapshot_id: str) -> bool:
        return self._run_storage.has_job_snapshot(snapshot_id)

    @traced
    def get_execution_plan_snapshot(self, snapshot_id: str) -> "ExecutionPlanSnapshot":
        return self._run_storage.get_execution_plan_snapshot(snapshot_id)

    @traced
    def add_snapshot(
        self,
        snapshot: Union["JobSnap", "ExecutionPlanSnapshot"],
    ) -> None:
        return self._run_storage.add_snapshot(snapshot)

    def wipe(self) -> None:
        self._run_storage.wipe()
        self._event_storage.wipe()

    # asset storage methods are now in AssetMixin

    # directories

    def __enter__(self) -> Self:
        from dagster._core.storage.defs_state.base import set_defs_state_storage

        self._exit_stack = ExitStack()
        self._exit_stack.enter_context(set_defs_state_storage(self.defs_state_storage))
        return self

    def __exit__(
        self,
        _exception_type: type[BaseException] | None,
        _exception_value: BaseException | None,
        _traceback: TracebackType | None,
    ) -> None:
        self.dispose()
        self._exit_stack.close()

    # backfill

    @property
    def should_start_background_run_thread(self) -> bool:
        """Gate on a feature to start a thread that monitors for if the run should be canceled."""
        return False

    def inject_env_vars(self, location_name: str | None) -> None:
        if not self._secrets_loader:
            return

        new_env = self._secrets_loader.get_secrets_for_environment(location_name)
        for k, v in new_env.items():
            os.environ[k] = v

    def backfill_log_storage_enabled(self) -> bool:
        return False

    def da_request_backfills(self) -> bool:
        return False

    def dagster_observe_supported(self) -> bool:
        return False



from collections.abc import Iterable, Sequence

from dagster_shared.serdes.serdes import deserialize_value

import dagster._check as check
from dagster._core.definitions.partitions.context import require_full_partition_loading_context
from dagster._core.definitions.partitions.definition.partitions_definition import (
    PartitionsDefinition,
)
from dagster._core.definitions.partitions.partition_key_range import PartitionKeyRange
from dagster._core.definitions.partitions.snap import PartitionsSnap
from dagster._core.definitions.partitions.subset.default import DefaultPartitionsSubset
from dagster._core.definitions.partitions.subset.partitions_subset import PartitionsSubset
from dagster._record import record
from dagster._serdes import serialize_value, whitelist_for_serdes


@whitelist_for_serdes
@record
class KeyRangesPartitionsSubset(PartitionsSubset):
    key_ranges: Sequence[PartitionKeyRange]
    partitions_snap: PartitionsSnap

    @require_full_partition_loading_context
    def get_partition_keys_not_in_subset(
        self, partitions_def: PartitionsDefinition
    ) -> Iterable[str]:
        partition_key_set = set(self.get_partition_keys())
        return [key for key in partitions_def.get_partition_keys() if key not in partition_key_set]

    @property
    def is_empty(self) -> bool:
        return len(self.key_ranges) == 0

    @property
    def partitions_definition(self) -> PartitionsDefinition:
        return self.partitions_snap.get_partitions_definition()

    @property
    def partition_key_ranges(self) -> Sequence[PartitionKeyRange]:
        return self.key_ranges

    def get_partition_key_ranges(
        self, partitions_def: PartitionsDefinition
    ) -> Sequence[PartitionKeyRange]:
        return self.key_ranges

    @require_full_partition_loading_context
    def get_partition_keys(self) -> Iterable[str]:
        keys = []
        for partition_key_range in self.key_ranges:
            keys.extend(self.partitions_definition.get_partition_keys_in_range(partition_key_range))
        return keys

    @require_full_partition_loading_context
    def with_partition_keys(self, partition_keys: Iterable[str]) -> "PartitionsSubset[str]":
        return DefaultPartitionsSubset({*self.get_partition_keys(), *partition_keys})

    def serialize(self) -> str:
        return serialize_value(self)

    @classmethod
    def from_serialized(
        cls, partitions_def: PartitionsDefinition, serialized: str
    ) -> "PartitionsSubset":
        return deserialize_value(serialized, KeyRangesPartitionsSubset)

    @classmethod
    def can_deserialize(
        cls,
        partitions_def: PartitionsDefinition,
        serialized: str,
        serialized_partitions_def_unique_id: str | None,
        serialized_partitions_def_class_name: str | None,
    ) -> bool:
        return True

    @require_full_partition_loading_context
    def __len__(self) -> int:
        return sum(
            [
                len(self.partitions_definition.get_partition_keys_in_range(partition_key_range))
                for partition_key_range in self.key_ranges
            ]
        )

    @require_full_partition_loading_context
    def __contains__(self, value) -> bool:
        return value in self.get_partition_keys()

    @classmethod
    def create_empty_subset(
        cls, partitions_def: PartitionsDefinition | None = None
    ) -> "KeyRangesPartitionsSubset":
        return KeyRangesPartitionsSubset(
            key_ranges=[], partitions_snap=PartitionsSnap.from_def(check.not_none(partitions_def))
        )

    def empty_subset(
        self,
    ) -> "KeyRangesPartitionsSubset":
        return KeyRangesPartitionsSubset(key_ranges=[], partitions_snap=self.partitions_snap)



from collections.abc import Iterable, Mapping
from functools import cached_property
from typing import TYPE_CHECKING, AbstractSet, Optional, Sequence  # noqa: UP035

from dagster_shared.record import record

import dagster._check as check
from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_key import AssetKey, EntityKey
from dagster._core.definitions.dependency import NodeHandle, NodeInputHandle, NodeOutputHandle
from dagster._core.definitions.graph_definition import GraphDefinition

if TYPE_CHECKING:
    from dagster._core.definitions.assets.definition.assets_definition import AssetsDefinition
    from dagster._core.definitions.assets.graph.asset_graph import AssetGraph, AssetNode


@record(checked=False)
class AssetLayerData:
    """Data that relates asset-level information to a node in the execution graph."""

    node_handle: NodeHandle
    assets_def: "AssetsDefinition"


@record(checked=False)
class AssetLayer:
    """Stores all of the asset-related information for a Dagster job. Maps execution node data
    to the backing AssetsDefinition, with a variety of convenience methods for common operations.
    """

    data: Sequence[AssetLayerData]
    asset_graph: "AssetGraph"
    # these are set when we have knowledge that a job corresponds to a set
    # of asset keys but don't have any AssetsDefinitions / nodes to map them
    # to. this can happen when we're representing a job that executes in an
    # external system but we don't have a good representation of the nodes
    # contained within it.
    external_job_asset_keys: AbstractSet[AssetKey] = set()
    # these are set when there is additional information that adds mapping
    # information from asset keys to input handles that is not captured by
    # the set of AssetsDefinitions that we have available to us. this can
    # happen when a source asset is used as input to a regular graph.
    mapped_source_asset_keys_by_input_handle: Mapping[NodeInputHandle, AssetKey] = {}

    @property
    def executable_asset_keys(self) -> Iterable[AssetKey]:
        return self.asset_graph.executable_asset_keys

    @cached_property
    def selected_asset_keys(self) -> AbstractSet[AssetKey]:
        return set().union(*(d.assets_def.keys for d in self.data))

    @cached_property
    def _data_by_handle(self) -> Mapping[NodeHandle, AssetLayerData]:
        return {data.node_handle: data for data in self.data}

    @cached_property
    def _data_by_key(self) -> Mapping[EntityKey, AssetLayerData]:
        mapping: dict[EntityKey, AssetLayerData] = {}
        for data in self.data:
            for key in [
                *data.assets_def.keys,
                *(spec.key for spec in data.assets_def.node_check_specs_by_output_name.values()),
            ]:
                mapping[key] = data
        return mapping

    @cached_property
    def _asset_keys_by_inner_input_handle(self) -> Mapping[NodeInputHandle, AssetKey]:
        """A mapping from the inner (op) input handles to the asset key that they
        correspond to.

        This is done with a cached property because there is no way of mapping from
        an inner node handle to the outer graph input handle that it came from, so
        we instead do this as a single bulk operation.
        """
        # seed the dictionary with any additional node inputs that were provided.
        # this is provided when source assets are used as inputs to a regular graph.
        mapping: dict[NodeInputHandle, AssetKey] = dict(
            self.mapped_source_asset_keys_by_input_handle
        )

        for data in self.data:
            assets_def = data.assets_def
            node_def = check.not_none(assets_def.computation).full_node_def
            outer_node_handle = data.node_handle

            # INPUTS: record mapping from outer inputs to inner inputs
            for outer_input_name, key in assets_def.node_keys_by_input_name.items():
                outer_input_handle = NodeInputHandle(
                    node_handle=outer_node_handle, input_name=outer_input_name
                )
                mapping[outer_input_handle] = key
                inner_input_handles = node_def.resolve_input_to_destinations(outer_input_handle)
                for inner_input_handle in inner_input_handles:
                    mapping[inner_input_handle] = key

            # OUTPUTS: record mapping from outer outputs to all inputs that consume them
            for outer_output_name, key in assets_def.node_keys_by_output_name.items():
                # find inputs that consume this output
                input_handles = node_def.resolve_output_to_destinations(
                    output_name=outer_output_name, handle=outer_node_handle
                )
                for outer_input_handle in input_handles:
                    mapping[outer_input_handle] = key
                    # ensure that we map from connected graph inputs to the inner inputs they connect to
                    if isinstance(node_def, GraphDefinition):
                        inner_node = node_def.get_node(outer_input_handle.node_handle.pop())
                        inner_input_handles = inner_node.definition.resolve_input_to_destinations(
                            outer_input_handle
                        )
                        for inner_input_handle in inner_input_handles:
                            mapping[inner_input_handle] = key

        return mapping

    @staticmethod
    def from_mapping(
        assets_defs_by_outer_node_handle: Mapping[NodeHandle, "AssetsDefinition"],
        asset_graph: "AssetGraph",
    ) -> "AssetLayer":
        return AssetLayer(
            data=[
                AssetLayerData(node_handle=node_handle, assets_def=assets_def)
                for node_handle, assets_def in assets_defs_by_outer_node_handle.items()
            ],
            asset_graph=asset_graph,
        )

    @staticmethod
    def for_external_job(asset_keys: Iterable[AssetKey]) -> "AssetLayer":
        from dagster._core.definitions.assets.graph.asset_graph import AssetGraph

        return AssetLayer(
            data=[], asset_graph=AssetGraph.from_assets([]), external_job_asset_keys=set(asset_keys)
        )

    def get(self, asset_key: AssetKey) -> "AssetNode":
        return self.asset_graph.get(asset_key)

    def has(self, asset_key: AssetKey) -> bool:
        return self.asset_graph.has(asset_key)

    def _maybe_get_data(self, ptr: NodeHandle | EntityKey) -> AssetLayerData | None:
        if isinstance(ptr, NodeHandle):
            # we need to ensure that the node handle we use for the lookup is the **outer**
            # node handle when in the context of a graph asset.
            while ptr.parent is not None:
                ptr = ptr.parent
            return self._data_by_handle.get(ptr)
        else:
            return self._data_by_key.get(ptr)

    def _get_data(self, ptr: NodeHandle | EntityKey) -> AssetLayerData:
        return check.not_none(self._maybe_get_data(ptr), f"Could not find AssetLayerData for {ptr}")

    def get_op_output_handle(self, key: EntityKey) -> NodeOutputHandle:
        data = self._get_data(key)
        computation = check.not_none(data.assets_def.computation)
        # the outer node handle that we store may refer to a graph definition,
        # so ensure that we resolve the output to the actual op output
        output_name = computation.output_names_by_entity_key[key]
        inner_output_def, inner_node_handle = computation.node_def.resolve_output_to_origin(
            output_name, data.node_handle
        )
        return NodeOutputHandle(node_handle=inner_node_handle, output_name=inner_output_def.name)

    def get_assets_def_for_node(self, node_handle: NodeHandle) -> Optional["AssetsDefinition"]:
        data = self._maybe_get_data(node_handle)
        return data.assets_def if data else None

    def get_selected_entity_keys_for_node(self, node_handle: NodeHandle) -> AbstractSet[EntityKey]:
        data = self._maybe_get_data(node_handle)
        return data.assets_def.asset_and_check_keys if data else set()

    def get_asset_key_for_node(self, node_handle: NodeHandle) -> AssetKey:
        return self._get_data(node_handle).assets_def.key

    def get_op_output_name(self, key: EntityKey) -> str:
        return self.get_op_output_handle(key).output_name

    def get_node_input_name(self, node_handle: NodeHandle, key: AssetKey) -> str | None:
        data = self._maybe_get_data(node_handle)
        return data.assets_def.input_names_by_node_key.get(key) if data else None

    def get_entity_key_for_node_output(
        self, inner_node_handle: NodeHandle, output_name: str
    ) -> EntityKey | None:
        data = self._maybe_get_data(inner_node_handle)
        if data is None:
            return None
        computation = check.not_none(data.assets_def.computation)
        return computation.entity_keys_by_op_output_handle.get(
            NodeOutputHandle(node_handle=inner_node_handle.pop(), output_name=output_name)
        )

    def get_asset_key_for_node_output(
        self, node_handle: NodeHandle, output_name: str
    ) -> AssetKey | None:
        key = self.get_entity_key_for_node_output(node_handle, output_name)
        return key if isinstance(key, AssetKey) else None

    def get_asset_check_key_for_node_output(
        self, node_handle: NodeHandle, output_name: str
    ) -> AssetCheckKey | None:
        key = self.get_entity_key_for_node_output(node_handle, output_name)
        return key if isinstance(key, AssetCheckKey) else None

    def get_asset_key_for_node_input(
        self, inner_node_handle: NodeHandle, input_name: str
    ) -> AssetKey | None:
        node_input_handle = NodeInputHandle(node_handle=inner_node_handle, input_name=input_name)
        return self._asset_keys_by_inner_input_handle.get(node_input_handle)

    def downstream_dep_assets_and_checks(
        self, node_handle: NodeHandle, output_name: str
    ) -> set[EntityKey]:
        """Given the node handle of an op within a graph-backed asset and an output name,
        returns the asset keys dependent on that output.

        For example, for the following asset:

        @op(out={"out_1": Out(is_required=False), "out_2": Out(is_required=False)})
        def two_outputs_op(context):
            return 1, 1


        @op
        def add_one(x):
            return x + 1


        @graph(out={"asset_one": GraphOut(), "asset_two": GraphOut()})
        def my_graph():
            out_1, out_2 = two_outputs_op()
            return {"asset_one": out_1, "asset_two": add_one(out_2)}

        two_assets = AssetsDefinition.from_graph(my_graph)

        Calling downstream_dep_assets with the node handle of two_outputs_op will return:
        - {AssetKey("asset_one")} if output_name="out_1"
        - {AssetKey("asset_two")} if output_name="out_2"

        Calling downstream_dep_assets with node handle add_one will return:
        - {AssetKey("asset_two")} if output_name="result"
        """
        assets_def = self._get_data(node_handle).assets_def
        return {
            key
            for key in check.not_none(assets_def.computation).entity_keys_by_dep_op_output_handle[
                NodeOutputHandle(node_handle=node_handle.pop(), output_name=output_name)
            ]
        }

    def upstream_dep_op_handles(self, asset_key: AssetKey) -> AbstractSet[NodeHandle]:
        data = self._get_data(asset_key)
        computation = check.not_none(data.assets_def.computation)
        op_handles_in_assets_def = computation.dep_op_handles_by_entity_key[asset_key]
        outer_node_handle = NodeHandle(data.node_handle.name, parent=None)
        return {outer_node_handle.with_child(op_handle) for op_handle in op_handles_in_assets_def}



from dagster.components.core.defs_module import DefsFolderComponent as DefsFolderComponent
from dagster.components.lib.definitions_component import (
    DefinitionsComponent as DefinitionsComponent,
)

# These just modify the exsiting Dagster decorators
from dagster.components.lib.shim_components.asset import asset as asset
from dagster.components.lib.shim_components.asset_check import asset_check as asset_check
from dagster.components.lib.shim_components.job import job as job
from dagster.components.lib.shim_components.multi_asset import multi_asset as multi_asset
from dagster.components.lib.shim_components.schedule import schedule as schedule
from dagster.components.lib.shim_components.sensor import sensor as sensor



from dagster._core.definitions.partitions.partitioned_config.dynamic import (
    dynamic_partitioned_config as dynamic_partitioned_config,
)
from dagster._core.definitions.partitions.partitioned_config.partitioned_config import (
    PartitionConfigFn as PartitionConfigFn,
    PartitionedConfig as PartitionedConfig,
    partitioned_config as partitioned_config,
)
from dagster._core.definitions.partitions.partitioned_config.static import (
    static_partitioned_config as static_partitioned_config,
)
from dagster._core.definitions.partitions.partitioned_config.time_window import (
    daily_partitioned_config as daily_partitioned_config,
    hourly_partitioned_config as hourly_partitioned_config,
    monthly_partitioned_config as monthly_partitioned_config,
    weekly_partitioned_config as weekly_partitioned_config,
)



from collections.abc import Sequence
from typing import NamedTuple, cast

import dagster._check as check
from dagster._annotations import beta, public
from dagster._core.definitions.partitions.definition.multi import MultiPartitionsDefinition
from dagster._core.definitions.partitions.definition.partitions_definition import (
    PartitionsDefinition,
)
from dagster._core.definitions.partitions.definition.time_window import (
    TimeWindowPartitionsDefinition,
)
from dagster._core.definitions.partitions.mapping.identity import IdentityPartitionMapping
from dagster._core.definitions.partitions.mapping.multi.base import (
    BaseMultiPartitionMapping,
    DimensionDependency,
)
from dagster._core.definitions.partitions.mapping.partition_mapping import PartitionMapping
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._serdes import whitelist_for_serdes


class InferSingleToMultiDimensionDepsResult(
    NamedTuple(
        "_InferSingleToMultiDimensionDepsResult",
        [
            ("can_infer", bool),
            ("inference_failure_reason", str | None),
            ("dimension_dependency", DimensionDependency | None),
        ],
    )
):
    def __new__(
        cls,
        can_infer: bool,
        inference_failure_reason: str | None = None,
        dimension_dependency: DimensionDependency | None = None,
    ):
        if can_infer and dimension_dependency is None:
            check.failed("dimension_dependency must be provided if can_infer is True")
        if not can_infer and inference_failure_reason is None:
            check.failed("inference_failure_reason must be provided if can_infer is False")

        return super().__new__(
            cls,
            can_infer,
            inference_failure_reason,
            dimension_dependency,
        )


def get_infer_single_to_multi_dimension_deps_result(
    upstream_partitions_def: PartitionsDefinition,
    downstream_partitions_def: PartitionsDefinition,
    partition_dimension_name: str | None = None,
) -> InferSingleToMultiDimensionDepsResult:
    from dagster._core.definitions.partitions.mapping import TimeWindowPartitionMapping

    upstream_is_multipartitioned = isinstance(upstream_partitions_def, MultiPartitionsDefinition)

    multipartitions_defs = [
        partitions_def
        for partitions_def in [upstream_partitions_def, downstream_partitions_def]
        if isinstance(partitions_def, MultiPartitionsDefinition)
    ]
    if len(multipartitions_defs) != 1:
        return InferSingleToMultiDimensionDepsResult(
            False,
            "Can only use MultiToSingleDimensionPartitionMapping when upstream asset is"
            " multipartitioned and the downstream asset is single dimensional, or vice versa."
            f" Instead received {len(multipartitions_defs)} multi-partitioned assets.",
        )

    multipartitions_def = cast("MultiPartitionsDefinition", next(iter(multipartitions_defs)))

    single_dimension_partitions_def = next(
        iter(
            {
                upstream_partitions_def,
                downstream_partitions_def,
            }
            - set(multipartitions_defs)
        )
    )

    filtered_multipartition_dims = (
        multipartitions_def.partitions_defs
        if partition_dimension_name is None
        else [
            dim
            for dim in multipartitions_def.partitions_defs
            if dim.name == partition_dimension_name
        ]
    )

    if partition_dimension_name:
        if len(filtered_multipartition_dims) != 1:
            return InferSingleToMultiDimensionDepsResult(
                False,
                f"Provided partition dimension name {partition_dimension_name} not found in"
                f" multipartitions definition {multipartitions_def}.",
            )

    matching_dimension_defs = [
        dimension_def
        for dimension_def in filtered_multipartition_dims
        if dimension_def.partitions_def == single_dimension_partitions_def
    ]

    if len(matching_dimension_defs) == 1:
        return InferSingleToMultiDimensionDepsResult(
            True,
            dimension_dependency=DimensionDependency(
                IdentityPartitionMapping(),
                upstream_dimension_name=(
                    matching_dimension_defs[0].name if upstream_is_multipartitioned else None
                ),
                downstream_dimension_name=(
                    matching_dimension_defs[0].name if not upstream_is_multipartitioned else None
                ),
            ),
        )
    elif len(matching_dimension_defs) > 1:
        return InferSingleToMultiDimensionDepsResult(
            False,
            "partition dimension name must be specified when multiple dimensions of the"
            " MultiPartitionsDefinition match the single dimension partitions def",
        )

    time_dimensions = [
        dimension_def
        for dimension_def in filtered_multipartition_dims
        if isinstance(dimension_def.partitions_def, TimeWindowPartitionsDefinition)
    ]

    if len(time_dimensions) == 1 and isinstance(
        single_dimension_partitions_def, TimeWindowPartitionsDefinition
    ):
        return InferSingleToMultiDimensionDepsResult(
            True,
            dimension_dependency=DimensionDependency(
                TimeWindowPartitionMapping(),
                upstream_dimension_name=(
                    time_dimensions[0].name if upstream_is_multipartitioned else None
                ),
                downstream_dimension_name=(
                    time_dimensions[0].name if not upstream_is_multipartitioned else None
                ),
            ),
        )

    return InferSingleToMultiDimensionDepsResult(
        False,
        "MultiToSingleDimensionPartitionMapping can only be used when: \n(a) The single dimensional"
        " partitions definition is a dimension of the MultiPartitionsDefinition.\n(b) The single"
        " dimensional partitions definition is a TimeWindowPartitionsDefinition and the"
        " MultiPartitionsDefinition has a single time dimension.",
    )


@public
@beta
@whitelist_for_serdes
class MultiToSingleDimensionPartitionMapping(
    BaseMultiPartitionMapping,
    PartitionMapping,
    NamedTuple(
        "_MultiToSingleDimensionPartitionMapping", [("partition_dimension_name", str | None)]
    ),
):
    """Defines a correspondence between an single-dimensional partitions definition
    and a MultiPartitionsDefinition. The single-dimensional partitions definition must be
    a dimension of the MultiPartitionsDefinition.

    This class handles the case where the upstream asset is multipartitioned and the
    downstream asset is single dimensional, and vice versa.

    For a partition key X, this partition mapping assumes that any multi-partition key with
    X in the selected dimension is a dependency.

    Args:
        partition_dimension_name (Optional[str]): The name of the partition dimension in the
            MultiPartitionsDefinition that matches the single-dimension partitions definition.
    """

    def __new__(cls, partition_dimension_name: str | None = None):
        return super().__new__(
            cls,
            partition_dimension_name=check.opt_str_param(
                partition_dimension_name, "partition_dimension_name"
            ),
        )

    @property
    def description(self) -> str:
        return (
            "Assumes that the single-dimension partitions definition is a dimension of the "
            "multi-partitions definition. For a single-dimension partition key X, any "
            "multi-partition key with X in the matching dimension is a dependency."
        )

    def validate_partition_mapping(
        self,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition | None,
    ):
        if not downstream_partitions_def:
            raise DagsterInvalidDefinitionError(
                "downstream_partitions_def must be provided for MultiToSingleDimensionPartitionMapping"
            )
        infer_mapping_result = get_infer_single_to_multi_dimension_deps_result(
            upstream_partitions_def, downstream_partitions_def
        )
        if not infer_mapping_result.can_infer:
            check.invariant(isinstance(infer_mapping_result.inference_failure_reason, str))
            check.failed(cast("str", infer_mapping_result.inference_failure_reason))

    def get_dimension_dependencies(
        self,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition,
    ) -> Sequence[DimensionDependency]:
        infer_mapping_result = get_infer_single_to_multi_dimension_deps_result(
            upstream_partitions_def, downstream_partitions_def
        )

        if not infer_mapping_result.can_infer:
            check.invariant(isinstance(infer_mapping_result.inference_failure_reason, str))
            check.failed(cast("str", infer_mapping_result.inference_failure_reason))

        return [cast("DimensionDependency", infer_mapping_result.dimension_dependency)]
